# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Unittests for the checkin module."""

from django.test import TransactionTestCase

from mock import Mock
from mock import patch

from pykeg.core import checkin
from pykeg.core import models
from pykeg.core.util import get_version


class CheckinTestCase(TransactionTestCase):

    def test_checkin(self):

        site = models.KegbotSite.get()
        site.registration_id = 'original-regid'
        site.save()

        version = get_version()

        with patch('requests.post') as mock_post:
            mock_post.return_value = mock_response = Mock()

            mock_response.status_code = 200
            mock_response.json.return_value = {
                'status': 'ok',
                'reg_id': 'new-regid'
            }
            checkin.checkin('http://example.com/checkin', 'test-product', 1.23)
            mock_post.assert_called_with('http://example.com/checkin',
                headers={'User-Agent': 'KegbotServer/%s' % version},
                data={
                    'reg_id': u'original-regid',
                    'product': 'test-product',
                    'version': version,
                },
                timeout=1.23)

        site = models.KegbotSite.get()
        self.assertEquals('new-regid', site.registration_id)



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Kegweb main views."""

from django.contrib import messages
from django.shortcuts import get_object_or_404
from django.shortcuts import render_to_response
from django.shortcuts import redirect
from django.template import RequestContext
from django.views.decorators.cache import cache_page
from django.views.generic.dates import ArchiveIndexView
from django.views.generic.dates import DateDetailView
from django.views.generic.dates import DayArchiveView
from django.views.generic.dates import MonthArchiveView
from django.views.generic.dates import YearArchiveView
from django.views.generic.list import ListView
from django.core.paginator import Paginator, EmptyPage, PageNotAnInteger

from pykeg.core import models
from pykeg.web.kegweb import forms

### main views


@cache_page(30)
def index(request):
    context = RequestContext(request)

    context['taps'] = models.KegTap.objects.all()
    context['events'] = models.SystemEvent.objects.timeline()[:20]
    sessions = models.DrinkingSession.objects.all().order_by('-id')[:10]
    context['sessions'] = sessions

    if sessions:
        last_session = sessions[0]
        context['most_recent_session'] = last_session
        if sessions and last_session.IsActiveNow():
            context['current_session'] = last_session

    return render_to_response('index.html', context_instance=context)


@cache_page(30)
def system_stats(request):
    stats = models.KegbotSite.get().get_stats()
    context = RequestContext(request, {
        'stats': stats,
    })

    top_drinkers = []
    for username, vol in stats.get('volume_by_drinker', {}).iteritems():
        try:
            user = models.User.objects.get(username=username)
        except models.User.DoesNotExist:
            continue  # should not happen
        top_drinkers.append((vol, user))
    top_drinkers.sort(reverse=True)

    largest_session_id = stats.get('largest_session', {}).get('session_id', None)
    if largest_session_id:
        try:
            context['largest_session'] = models.DrinkingSession.objects.get(pk=largest_session_id)
        except models.DrinkingSession.DoesNotExist:
            # Stats out of date.
            pass

    context['top_drinkers'] = top_drinkers[:10]

    return render_to_response('kegweb/system-stats.html', context_instance=context)


### object lists and detail (generic views)

def user_detail(request, username):
    user = get_object_or_404(models.User, username=username, is_active=True)
    stats = user.get_stats()
    drinks = user.drinks.all()

    context = RequestContext(request, {
        'drinks': drinks,
        'stats': stats,
        'drinker': user})

    largest_session_id = stats.get('largest_session', {}).get('session_id', None)
    if largest_session_id:
        context['largest_session'] = models.DrinkingSession.objects.get(pk=largest_session_id)

    return render_to_response('kegweb/drinker_detail.html', context_instance=context)


class KegListView(ListView):
    model = models.Keg
    template_name = 'kegweb/keg_list.html'
    context_object_name = 'kegs'
    paginate_by = 10

    def get_queryset(self):
        return models.Keg.objects.all().order_by('-id')


def fullscreen(request):
    context = RequestContext(request)
    taps = models.KegTap.objects.all()
    active_taps = [t for t in taps if t.current_keg]
    pages = [active_taps[i:i + 4] for i in range(0, len(active_taps), 4)]
    context['pages'] = pages

    return render_to_response('kegweb/fullscreen.html', context_instance=context)


@cache_page(30)
def keg_detail(request, keg_id):
    keg = get_object_or_404(models.Keg, id=keg_id)
    sessions = keg.get_sessions()
    last_session = sessions[:1]

    context = RequestContext(request, {
        'keg': keg,
        'stats': keg.get_stats(),
        'sessions': sessions,
        'last_session': last_session})
    return render_to_response('kegweb/keg_detail.html', context_instance=context)


def short_drink_detail(request, drink_id):
    return redirect('kb-drink', drink_id=str(drink_id), permanent=True)


def short_session_detail(request, session_id):
    session = get_object_or_404(models.DrinkingSession, id=session_id)
    url = session.get_absolute_url()
    return redirect(url, permanent=True)


def drink_detail(request, drink_id):
    drink = get_object_or_404(models.Drink, id=drink_id)
    context = RequestContext(request, {'drink': drink})

    can_delete = (request.user == drink.user) or request.user.is_staff

    if can_delete:
        picture_form = forms.DeletePictureForm(initial={'picture': drink.picture})
    else:
        picture_form = None

    if request.method == 'POST':
        if can_delete:
            picture_form = forms.DeletePictureForm(request.POST)
            if picture_form.is_valid():
                drink.picture.erase_and_delete()
                picture_form = None
                messages.success(request, 'Erased image.')
            else:
                messages.error(request, 'request not valid: ' + str(picture_form.errors))
        else:
            messages.error(request, 'No permission to delete picture.')
        return redirect('kb-drink', drink_id=str(drink_id))

    context['picture_form'] = picture_form
    return render_to_response('kegweb/drink_detail.html', context_instance=context)


def drinker_sessions(request, username):
    user = get_object_or_404(models.User, username=username, is_active=True)
    stats = user.get_stats()
    drinks = user.drinks.all()

    chunks = models.Stats.objects.filter(
        user=user, keg__isnull=True, session__isnull=False, is_first=True
    ).order_by('-id').select_related('session')

    paginator = Paginator(chunks, 5)

    page = request.GET.get('page')
    try:
        chunks = paginator.page(page)
    except PageNotAnInteger:
        chunks = paginator.page(1)
    except EmptyPage:
        chunks = paginator.page(paginator.num_pages)

    context = RequestContext(request, {
        'drinks': drinks,
        'chunks': chunks,
        'stats': stats,
        'drinker': user})

    return render_to_response('kegweb/drinker_sessions.html', context_instance=context)


def keg_sessions(request, keg_id):
    keg = get_object_or_404(models.Keg, id=keg_id)
    sessions = keg.get_sessions()

    paginator = Paginator(sessions, 5)

    page = request.GET.get('page')
    try:
        sessions = paginator.page(page)
    except PageNotAnInteger:
        sessions = paginator.page(1)
    except EmptyPage:
        sessions = paginator.page(paginator.num_pages)

    context = RequestContext(request, {
        'keg': keg,
        'stats': keg.get_stats(),
        'sessions': sessions})
    return render_to_response('kegweb/keg_sessions.html', context_instance=context)


class SessionArchiveIndexView(ArchiveIndexView):
    model = models.DrinkingSession
    date_field = 'start_time'
    template_name = 'kegweb/drinkingsession_archive.html'
    context_object_name = 'sessions'
    paginate_by = 20


class SessionYearArchiveView(YearArchiveView):
    model = models.DrinkingSession
    date_field = 'start_time'
    template_name = 'kegweb/drinkingsession_archive_year.html'
    make_object_list = True
    context_object_name = 'sessions'
    paginate_by = 20


class SessionMonthArchiveView(MonthArchiveView):
    model = models.DrinkingSession
    date_field = 'start_time'
    template_name = 'kegweb/drinkingsession_archive_month.html'
    make_object_list = True
    context_object_name = 'sessions'
    paginate_by = 20


class SessionDayArchiveView(DayArchiveView):
    model = models.DrinkingSession
    date_field = 'start_time'
    template_name = 'kegweb/drinkingsession_archive_day.html'
    make_object_list = True
    context_object_name = 'sessions'
    paginate_by = 20


class SessionDateDetailView(DateDetailView):
    model = models.DrinkingSession
    date_field = 'start_time'
    template_name = 'kegweb/session_detail.html'
    context_object_name = 'session'

    def get_context_data(self, **kwargs):
        """Adds `stats` to the context."""
        ret = super(SessionDateDetailView, self).get_context_data(**kwargs)
        stats = ret[self.context_object_name].get_stats()
        ret['stats'] = stats
        ret['kegs'] = [models.Keg.objects.get(pk=pk) for pk in stats.get('keg_ids', [])]
        return ret



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

from django.conf import settings
from django.conf.urls import include
from django.conf.urls import patterns
from django.conf.urls import url
from django.contrib import admin
from django.contrib.staticfiles.urls import staticfiles_urlpatterns
from django.views.generic.base import RedirectView

admin.autodiscover()

urlpatterns = patterns('',
    ### api
    (r'^api/(v1/)?', include('pykeg.web.api.urls')),

    ### kegbot account
    (r'^account/', include('pykeg.web.account.urls')),

    ### auth account
    (r'^accounts/', include('pykeg.web.kbregistration.urls')),

    ### kegadmin
    (r'^kegadmin/', include('pykeg.web.kegadmin.urls')),

    ### Shortcuts
    (r'^link/?$', RedirectView.as_view(pattern_name='kegadmin-link-device'))
)

if 'pykeg.web.setup_wizard' in settings.INSTALLED_APPS:
    urlpatterns += patterns('',
        (r'^setup/', include('pykeg.web.setup_wizard.urls')),
    )

if settings.DEBUG:
    urlpatterns += staticfiles_urlpatterns()
    urlpatterns += patterns('',
      url(r'^media/(?P<path>.*)$', 'django.views.static.serve', {'document_root': settings.MEDIA_ROOT, }),
    )

if settings.KEGBOT_ENABLE_ADMIN:
    urlpatterns += patterns('',
      (r'^admin/', include(admin.site.urls)),
    )

if settings.DEMO_MODE:
    urlpatterns += patterns('',
      (r'^demo/', include('pykeg.contrib.demomode.urls')),
    )

### main kegweb urls
urlpatterns += patterns('',
  (r'^', include('pykeg.web.kegweb.urls')),
)



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys

from django.core.management.base import BaseCommand
from django.core.management.base import CommandError
from pykeg.backup import backup


class Command(BaseCommand):
    help = u'Restores a zipfile backup of the current Kegbot system.'
    args = u'<zipfile>'

    def handle(self, *args, **options):
        if len(args) != 1:
            raise CommandError('Must provide a zipfile path')
        backup_path = os.path.normpath(os.path.expanduser(args[0]))
        if not os.path.exists(backup_path):
            raise CommandError('Archive does not exist: {}'.format(backup_path))

        try:
            backup.restore(backup_path)
        except backup.AlreadyInstalledError:
            sys.stderr.write('Error: Kegbot is already installed, run `kegbot erase` first.')
            sys.stderr.write('\n')
            sys.exit(1)
        except backup.BackupError as e:
            sys.stderr.write('Error: ')
            sys.stderr.write(e.message)
            sys.stderr.write('\n')
            sys.exit(1)



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Unittests for pykeg.backup.backup"""

import difflib
import os
import sys
import shutil
import tempfile

from pykeg.core import defaults
from pykeg.core import models

from django.core.files.storage import FileSystemStorage
from django.test import TransactionTestCase

from . import backup
from pykeg.core.testutils import make_datetime
from pykeg.core.util import get_version

from kegbot.util import kbjson


def run(cmd, args=[]):
    cmdname = cmd.__module__.split('.')[-1]
    cmd.run_from_argv([sys.argv[0], cmdname] + args)


class BackupTestCase(TransactionTestCase):
    def setUp(self):
        self.temp_storage_location = tempfile.mkdtemp(dir=os.environ.get('DJANGO_TEST_TEMP_DIR'))
        self.storage = FileSystemStorage(location=self.temp_storage_location)
        self.now = make_datetime(2014, 4, 1)

    def tearDown(self):
        shutil.rmtree(self.temp_storage_location)

    def assertMetadata(self, backup_dir, when=None, site_name='My Kegbot',
            num_media_files=0):
        when = when or self.now

        backup.verify_backup_directory(backup_dir)
        metadata_file = os.path.join(backup_dir, backup.METADATA_FILENAME)
        metadata_json = kbjson.loads(open(metadata_file).read())

        self.assertEqual(when, metadata_json[backup.META_CREATED_TIME])
        self.assertEquals(site_name, metadata_json[backup.META_SERVER_NAME])
        self.assertEquals(num_media_files, metadata_json[backup.META_NUM_MEDIA_FILES])
        self.assertEquals(get_version(), metadata_json[backup.META_SERVER_VERSION])

    def test_create_backup_tree_empty_site(self):
        # Create a backup of an empty system.
        backup_dir = backup.create_backup_tree(self.now, storage=self.storage)

        try:
            self.assertMetadata(backup_dir, site_name='kegbot')
        finally:
            shutil.rmtree(backup_dir)

    def test_create_backup_tree_with_defaults(self):
        defaults.set_defaults(set_is_setup=True)

        site = models.KegbotSite.get()
        site.title = 'Kegbot Test 2'
        site.save()

        backup_dir = backup.create_backup_tree(self.now, storage=self.storage)

        try:
            self.assertMetadata(backup_dir, site_name='Kegbot Test 2')
        finally:
            shutil.rmtree(backup_dir)

    def test_restore_needs_erase(self):
        defaults.set_defaults(set_is_setup=True)
        backup_dir = backup.create_backup_tree(self.now, storage=self.storage)

        try:
            # Restore must fail when something is already installed.
            self.assertRaises(backup.AlreadyInstalledError, backup.restore_from_directory,
                    backup_dir)

            # Erase and restore.
            backup.erase(self.storage)
            backup.restore_from_directory(backup_dir, self.storage)
        finally:
            shutil.rmtree(backup_dir)

    def test_backup_contents_same(self):
        defaults.set_defaults(set_is_setup=True)
        backup_dir = backup.create_backup_tree(self.now, storage=self.storage)

        try:
            # Verify recursive_diff works with no diffs.
            self.recursive_diff(backup_dir, backup_dir)

            second_backup_dir = backup.create_backup_tree(self.now, storage=self.storage)
            try:
                self.recursive_diff(backup_dir, second_backup_dir)
            finally:
                shutil.rmtree(second_backup_dir)
        finally:
            shutil.rmtree(backup_dir)

    def recursive_diff(self, dir1, dir2):
        dir1_files = set()
        dir2_files = set()

        for dirname, subdirs, files in os.walk(dir1):
            for filename in files:
                full_path = os.path.join(dirname, filename)
                rel_path = os.path.relpath(full_path, dir1)
                dir1_files.add(rel_path)

        for dirname, subdirs, files in os.walk(dir2):
            for filename in files:
                full_path = os.path.join(dirname, filename)
                rel_path = os.path.relpath(full_path, dir2)
                dir2_files.add(rel_path)

        self.assertEqual(dir1_files, dir2_files)

        for relfile in dir1_files:
            f1_full = os.path.join(dir1, relfile)
            f2_full = os.path.join(dir2, relfile)
            f1 = open(f1_full).read()
            f2 = open(f2_full).read()
            if f1 != f2:
                message = 'Files not equal: "{}" and "{}" differ.'.format(f1_full, f2_full)
                message += '\n' + ''.join(difflib.ndiff(f1.splitlines(True), f2.splitlines(True)))
                self.fail(message)



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Postgres-specific database backup/restore implementation."""

import logging
import os
import subprocess

from django.conf import settings

logger = logging.getLogger(__name__)

DEFAULT_DB = 'default'

# Common command-line arguments
PARAMS = {
    'db': settings.DATABASES[DEFAULT_DB].get('NAME'),
    'user': settings.DATABASES[DEFAULT_DB].get('USER'),
    'password': settings.DATABASES[DEFAULT_DB].get('PASSWORD'),
    'host': settings.DATABASES[DEFAULT_DB].get('HOST'),
    'port': settings.DATABASES[DEFAULT_DB].get('PORT'),
}

DEFAULT_ARGS = []
if PARAMS.get('user'):
    DEFAULT_ARGS.append('--username={}'.format(PARAMS['user']))
if PARAMS.get('host'):
    DEFAULT_ARGS.append('--host={}'.format(PARAMS['host']))
if PARAMS.get('port'):
    DEFAULT_ARGS.append('--port={}'.format(PARAMS['port']))

DEFAULT_ENV = dict(os.environ)
if PARAMS.get('password'):
    DEFAULT_ENV['PGPASSWORD'] = PARAMS['password']


def engine_name():
    return 'postgres'


def is_installed():
    args = ['psql'] + DEFAULT_ARGS
    args += ['-qt', '-c "select * from pg_tables where schemaname=\'public\';"', PARAMS['db']]
    cmd = ' '.join(args)
    logger.info(cmd)
    output = subprocess.check_output(cmd, env=DEFAULT_ENV, shell=True)
    return 'core_' in output


def dump(output_fd):
    args = ['pg_dump'] + DEFAULT_ARGS
    args.append(PARAMS['db'])
    cmd = ' '.join(args)
    logger.info(cmd)
    return subprocess.check_call(cmd, stdout=output_fd, env=DEFAULT_ENV, shell=True)


def restore(input_fd):
    args = ['psql'] + DEFAULT_ARGS
    args.append(PARAMS['db'])
    cmd = ' '.join(args)
    logger.info(cmd)
    return subprocess.check_call(cmd, stdin=input_fd, shell=True)


def erase():
    args = ['psql'] + DEFAULT_ARGS
    args += [PARAMS['db'], "-c 'drop schema public cascade; create schema public;'"]
    cmd = ' '.join(args)
    logger.info(cmd)
    subprocess.check_call(cmd, shell=True)



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from optparse import make_option
from distutils.version import StrictVersion
import sys

from pykeg.core.util import get_version_object

from django.core.management.base import BaseCommand

from django.contrib.staticfiles.management.commands import collectstatic
from django.core.management.commands import syncdb
from pykeg.core.management.commands import regen_stats

from pykeg.core import models
from pykeg.core import checkin


# Versions earlier than this cannot be upgraded. History:
#  v0.9.35 - migrations rebased to 0001
#  v1.1.1  - last release with South-based migrations
#  v1.2.0  - first Django 1.7 migrations
MINIMUM_INSTALLED_VERSION = StrictVersion('1.1.1')


def run(cmd, args=[]):
    cmdname = cmd.__module__.split('.')[-1]
    arg_str = ' '.join('%s' % a for a in args)
    print '--- Running command: %s %s' % (cmdname, arg_str)
    cmd.run_from_argv([sys.argv[0], cmdname] + args)


class Command(BaseCommand):
    help = u'Perform post-upgrade tasks.'
    option_list = BaseCommand.option_list + (
        make_option('--force', action='store_true', dest='force', default=False,
            help='Run even if installed version is up-to-date.'),
        make_option('--skip_static', action='store_true', dest='skip_static', default=False,
            help='Skip `kegbot collectstatic` during upgrade. (Not recommended.)'),
        make_option('--skip_stats', action='store_true', dest='skip_stats', default=False,
            help='Skip `kegbot regen_stats` during upgrade. (Not recommended.)'),
    )

    def handle(self, *args, **options):
        installed_version = models.KegbotSite.get_installed_version()
        app_version = get_version_object()
        force = options.get('force')

        if installed_version is None:
            print 'Kegbot is not installed; run setup-kegbot.py first.'
            sys.exit(1)

        if installed_version == app_version and not force:
            print 'Version {} already installed.'.format(installed_version)
            return

        if installed_version > app_version:
            print 'Installed version {} is newer than app version {}'.format(
                installed_version, app_version)
            sys.exit(1)

        if installed_version < MINIMUM_INSTALLED_VERSION:
            print ''
            print 'ERROR: This version of Kegbot can only upgrade systems running on version'
            print 'v{} or newer.  Please install Kegbot v{} and run `kegbot upgrade` again.'.format(
                MINIMUM_INSTALLED_VERSION, MINIMUM_INSTALLED_VERSION)
            print '(Existing version: {})'.format(installed_version)
            print ''
            print 'More help: https://github.com/Kegbot/kegbot-server/wiki/Upgrading-Old-Versions'
            print ''
            sys.exit(1)

        print 'Upgrading from {} to {}'.format(installed_version, app_version)
        self.do_version_upgrades(installed_version)

        run(syncdb.Command(), args=['--noinput', '-v', '0'])

        if not options.get('skip_stats'):
            run(regen_stats.Command())

        if not options.get('skip_static'):
            run(collectstatic.Command(), args=['--noinput'])

        site = models.KegbotSite.get()
        site.server_version = str(app_version)
        site.save()

        # Refresh any news (since we have a new version).
        try:
            checkin.checkin(timeout=5.0, quiet=True)
        except (checkin.CheckinError, Exception):
            pass

        print ''
        print 'Upgrade complete!'

    def do_version_upgrades(self, installed_version):
        if installed_version.version < (1, 2, 0):
            print 'Upgrading from v1.1.x'



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

import os
from django.conf import settings

from pykeg.core.management.commands.common import RunnerCommand


class Command(RunnerCommand):
    help = u'Runs background task queue workers.'
    pidfile_name = 'kegbot_run_workers.pid'

    def get_commands(self, options):
        default_log = stats_log = beat_log = ''
        logs_dir = options.get('logs_dir')

        if logs_dir:
            default_log = ' --logfile="{}"'.format(
                os.path.join(logs_dir, 'celery_default.log'))
            stats_log = ' --logfile="{}"'.format(
                os.path.join(logs_dir, 'celery_stats.log'))
            beat_log = ' --logfile="{}"'.format(
                os.path.join(logs_dir, 'celery_beat.log'))
        ret = []

        # Pluck out the app_name.  Typically this is 'pykeg', but may be
        # something else in embedded mode.
        app_name = settings.SETTINGS_MODULE.split('.')[0]
        base_cmd = 'celery -A {} worker -l info '.format(app_name)

        ret.append(('celery_default',
            base_cmd + '-Q default --hostname="default@%h"' + default_log))
        ret.append(('celery_stats',
            base_cmd + '-Q stats --concurrency=1 --hostname="stats@%h"' + stats_log))

        ret.append(('celery_beat',
            'celery -A {} beat --pidfile= -l info{}'.format(app_name, beat_log)))

        return ret



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from functools import wraps
from django.conf import settings
from django.contrib import messages
from django.contrib.auth import authenticate
from django.contrib.auth import login
from django.http import Http404
from django.shortcuts import redirect
from django.shortcuts import render_to_response
from django.template import RequestContext
from django.views.decorators.cache import never_cache

from pykeg.core import defaults
from pykeg.core import models

from .forms import AdminUserForm
from .forms import MiniSiteSettingsForm


def setup_view(f):
    """Decorator for setup views."""
    def new_function(*args, **kwargs):
        request = args[0]
        if not settings.DEBUG:
            raise Http404('Site is not in DEBUG mode.')
        if request.kbsite and request.kbsite.is_setup:
            raise Http404('Site is already setup, wizard disabled.')
        return f(*args, **kwargs)
    return wraps(f)(new_function)


@setup_view
@never_cache
def start(request):
    """ Shows the enable/disable hardware toggle. """
    context = RequestContext(request)

    if request.method == 'POST':
        if 'enable_sensing' in request.POST:
            request.session['enable_sensing'] = True
            return redirect('setup_accounts')
        elif 'disable_sensing' in request.POST:
            request.session['enable_sensing'] = False
            request.session['enable_users'] = False
            return redirect('setup_site_settings')
        else:
            messages.error(request, 'Unknown response.')

    return render_to_response('setup_wizard/start.html', context_instance=context)


@setup_view
@never_cache
def setup_accounts(request):
    """ Shows the enable/disable accounts toggle. """
    context = RequestContext(request)

    if request.method == 'POST':
        if 'enable_users' in request.POST:
            request.session['enable_users'] = True
            return redirect('setup_site_settings')
        elif 'disable_users' in request.POST:
            request.session['enable_users'] = False
            return redirect('setup_site_settings')
        else:
            messages.error(request, 'Unknown response.')

    return render_to_response('setup_wizard/accounts.html', context_instance=context)


@setup_view
@never_cache
def site_settings(request):
    context = RequestContext(request)

    if request.method == 'POST':
        form = MiniSiteSettingsForm(request.POST, instance=request.kbsite)
        if form.is_valid():
            form.save()
            messages.success(request, 'Settings saved!')
            return redirect('setup_admin')
    else:
        try:
            defaults.set_defaults()
        except defaults.AlreadyInstalledError:
            pass

        site = models.KegbotSite.get()
        site.enable_sensing = request.session['enable_sensing']
        site.enable_users = request.session['enable_users']
        site.save()
        form = MiniSiteSettingsForm(instance=site)
    context['form'] = form
    return render_to_response('setup_wizard/site_settings.html', context_instance=context)


@setup_view
@never_cache
def admin(request):
    context = RequestContext(request)
    form = AdminUserForm()
    if request.method == 'POST':
        form = AdminUserForm(request.POST)
        if form.is_valid():
            form.save()
            user = authenticate(username=form.cleaned_data.get('username'),
                password=form.cleaned_data.get('password'))
            if user:
                login(request, user)
            return redirect('setup_finish')
    context['form'] = form
    return render_to_response('setup_wizard/admin.html', context_instance=context)


@setup_view
@never_cache
def finish(request):
    context = RequestContext(request)
    if request.method == 'POST':
        request.kbsite.is_setup = True
        request.kbsite.save()
        messages.success(request, 'Tip: Install a new Keg in Admin: Taps')
        return redirect('kegadmin-main')
    return render_to_response('setup_wizard/finish.html', context_instance=context)



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Site backup and restore functions.

Backup file format is a zipfile with following contents:
    backup-name/         - top-level directory
       metadata.json     - backup metadata
       media/            - site media
       backup.sql        - backup sql data

The zipfile name takes the form <backup-name>-<sha1sum>.zip.

`backup-name` takes the form <site-title>-<YYMMDD-HHMMSS>, where
`site-title` is the slugified version of the current Kegbot site's
title.

Note that both the zipfile name and the backup file dirname are
informational and may change in a future version.

metadata.json consists of backup-related metadata:
  'created_time': ISO8601 timestamp at backup create
  'num_media_files': number of files in the 'media/' folder
  'server_version': server version at backup create time
  'server_name': server name at backup create time
"""

import hashlib
import logging
import os
import sys
import isodate
import tempfile
import shutil
import json
import zipfile

from pykeg.core.util import get_version
from kegbot.util import kbjson

from .exceptions import BackupError, InvalidBackup, AlreadyInstalledError

from django.conf import settings
from django.core.files.storage import default_storage
from django.db import connection
from django.db import transaction
from django.utils import timezone
from django.utils.text import slugify

logger = logging.getLogger(__name__)


# Whitelist of directories to include from storage backend.
MEDIA_WHITELIST = [
    u'pics'
]

# Metadata key names.
METADATA_FILENAME = 'metadata.json'
SQL_FILENAME = 'kegbot.sql'
META_CREATED_TIME = 'created_time'
META_NUM_MEDIA_FILES = 'num_media_files'
META_SERVER_VERSION = 'server_version'
META_SERVER_NAME = 'server_name'
META_BACKUP_FORMAT = 'backup_format'
META_DB_ENGINE = 'db_engine'

DEFAULT_DB = 'default'
BACKUPS_DIRNAME = 'backups'

# Current protocol.
BACKUP_FORMAT = 2


ENGINE_MYSQL = 'mysql'
ENGINE_POSTGRES = 'postgres'
ENGINE_UNKNOWN = 'unknown'

if 'mysql' in settings.DATABASES[DEFAULT_DB]['ENGINE']:
    from . import mysql as db_impl
elif 'postgres' in settings.DATABASES[DEFAULT_DB]['ENGINE']:
    from . import postgres as db_impl
else:
    from . import unknown_engine as db_impl


def get_title():
    with connection.cursor() as c:
        c.execute("SELECT title FROM core_kegbotsite WHERE name = 'default'")
        row = c.fetchone()
    if row:
        return row[0]
    return 'kegbot'


def read_metadata(zipfile):
    """Reads and returns metadata from a backup zipfile."""
    for info in zipfile.infolist():
        basename = os.path.basename(info.filename)
        if basename == METADATA_FILENAME:
            meta_file = zipfile.read(info.filename)
            return kbjson.loads(meta_file)
    return {}


@transaction.atomic
def create_backup_tree(date, storage, include_media=True):
    """Builds a complete backup in a temporary directory, return the path."""
    backup_dir = tempfile.mkdtemp()
    metadata = {}

    # Save databases.
    output_filename = os.path.join(backup_dir, SQL_FILENAME)
    with open(output_filename, 'w') as out_fd:
        db_impl.dump(out_fd)

    # Save stored media.
    metadata[META_NUM_MEDIA_FILES] = 0

    def add_files(storage, dirname, destdir):
        """Recursively copies all files in `dirname` to `destdir`."""
        subdirs, files = storage.listdir(dirname)
        for filename in files:
            full_filename = os.path.join(dirname, filename)
            output_filename = os.path.join(destdir, full_filename)
            output_dirname = os.path.dirname(output_filename)
            if not os.path.exists(output_dirname):
                os.makedirs(output_dirname)
            with storage.open(full_filename, 'r') as srcfile:
                with open(output_filename, 'w') as dstfile:
                    logger.debug('+++ Creating {}'.format(output_filename))
                    shutil.copyfileobj(srcfile, dstfile)
                    metadata[META_NUM_MEDIA_FILES] += 1
        for subdir in subdirs:
            add_files(storage, os.path.join((dirname, subdir)), destdir)

    if include_media:
        destdir = os.path.join(backup_dir, 'media')
        for media_dir in MEDIA_WHITELIST:
            if storage.exists(media_dir):
                add_files(storage, media_dir, destdir)
    else:
        logger.warning('Not including media.')

    # Store metadata file.
    metadata[META_SERVER_NAME] = get_title()
    metadata[META_SERVER_VERSION] = get_version()
    metadata[META_CREATED_TIME] = isodate.datetime_isoformat(date)
    metadata[META_DB_ENGINE] = db_impl.engine_name()
    metadata[META_BACKUP_FORMAT] = BACKUP_FORMAT
    metadata_filename = os.path.join(backup_dir, METADATA_FILENAME)
    with open(metadata_filename, 'w') as outfile:
        json.dump(metadata, outfile, sort_keys=True, indent=2)

    valid = False
    try:
        verify_backup_directory(backup_dir)
        valid = True
        return backup_dir
    finally:
        if not valid:
            shutil.rmtree(backup_dir)


def create_backup_zip(backup_dir, backup_name):
    """Creates a zipfile based on a tree created with `create_backup_tree()`."""
    zipfile_fileno, zipfile_path = tempfile.mkstemp()
    zipfile_fd = os.fdopen(zipfile_fileno, 'w')
    zf = zipfile.ZipFile(zipfile_fd, 'w')

    for dirname, subdirs, files in os.walk(backup_dir):
        for filename in files:
            full_path = os.path.join(dirname, filename)
            rel_path = os.path.relpath(full_path, backup_dir)
            archive_path = os.path.join(backup_name, rel_path)
            zf.write(full_path, archive_path)
            logger.debug('Added to zip: {}'.format(archive_path))

    zf.close()
    return zipfile_path


def backup(storage=default_storage, include_media=True):
    """Creates a new backup archive.

    Returns:
        A path to the stored zip file, in terms of the current storage class.
    """
    date = timezone.now()
    site_slug = slugify(get_title())
    date_str = date.strftime('%Y%m%d-%H%M%S')
    backup_name = '{slug}-{date}'.format(slug=site_slug, date=date_str)

    backup_dir = create_backup_tree(date=date, storage=storage, include_media=include_media)
    try:
        temp_zip = create_backup_zip(backup_dir, backup_name)
        try:
            # Append sha1sum.
            sha1 = hashlib.sha1()
            with open(temp_zip, 'rb') as f:
                for chunk in iter(lambda: f.read(2 ** 20), b''):
                    sha1.update(chunk)
            digest = sha1.hexdigest()
            saved_zip_name = os.path.join(BACKUPS_DIRNAME,
                '{}-{}.zip'.format(backup_name, digest))
            with open(temp_zip, 'r') as temp_zip_file:
                ret = storage.save(saved_zip_name, temp_zip_file)
                return ret
        finally:
            os.unlink(temp_zip)
    finally:
        shutil.rmtree(backup_dir)


def extract_backup(backup_file):
    assert sys.version_info[:3] >= (2, 7, 4), "Unsafe to extract ZIPs in this Python version"
    backup_dir = tempfile.mkdtemp()
    logger.debug('Extracting backup to {}'.format(backup_dir))
    zf = zipfile.ZipFile(backup_file, mode='r')
    zf.extractall(path=backup_dir)
    return backup_dir


def verify_backup_directory(backup_dir):
    metadata_file = os.path.join(backup_dir, METADATA_FILENAME)
    if not os.path.exists(metadata_file):
        raise InvalidBackup('Metadata file does not exist')

    if not os.path.exists(metadata_file):
        raise InvalidBackup('SQL dumpfile does not exist')

    metadata = kbjson.loads(open(metadata_file, 'r').read())
    format = metadata.get(META_BACKUP_FORMAT)
    if format != BACKUP_FORMAT:
        raise InvalidBackup('Unsupported backup format: {}'.format(format))

    return metadata


def restore_media(backup_dir, storage):
    media_dir = os.path.join(backup_dir, 'media')

    for dirname, subdirs, files in os.walk(media_dir):
        for filename in files:
            full_path = os.path.join(dirname, filename)
            rel_path = os.path.relpath(full_path, media_dir)
            with open(full_path, 'r') as data:
                logger.debug('+++ Restoring file {}'.format(rel_path))
                storage.save(rel_path, data)


def restore_from_directory(backup_dir, storage=default_storage):
    logger.info('Restoring from {} ...'.format(backup_dir))

    if db_impl.is_installed():
        raise AlreadyInstalledError('You must erase this system before restoring.')

    metadata = verify_backup_directory(backup_dir)
    current_engine = db_impl.engine_name()
    saved_engine = metadata[META_DB_ENGINE]
    if current_engine != saved_engine:
        raise BackupError('Current DB is {}; cannot restore from {}'.format(current_engine, db_impl))

    input_filename = os.path.join(backup_dir, SQL_FILENAME)
    with open(input_filename, 'r') as in_fd:
        db_impl.restore(in_fd)

    restore_media(backup_dir, storage)

    logger.info('Restore completed successfully.')


def restore(backup_file):
    """Restores from a backup zipfile or directory.

    The site must be erased prior to running restore.
    """
    if os.path.isdir(backup_file):
        restore_from_directory(backup_file)
        return

    logger.info('Restoring from {} ...'.format(backup_file))
    unzipped_dir = extract_backup(backup_file)
    try:
        dirs = os.listdir(unzipped_dir)
        if len(dirs) != 1:
            raise InvalidBackup('Expected exactly one directory in archive.')
        backup_dir = os.path.join(unzipped_dir, dirs[0])
        restore_from_directory(backup_dir)
    finally:
        shutil.rmtree(unzipped_dir)


def erase(storage=default_storage):
    """Erases ALL site data (database tables, media files)."""
    logger.info('Erasing tables ...')
    db_impl.erase()

    def delete_files(dirname):
        subdirs, files = storage.listdir(dirname)
        for filename in files:
            full_name = os.path.join(dirname, filename)
            logger.debug('Deleting file: {}'.format(full_name))
            storage.delete(full_name)
        for subdir in subdirs:
            delete_files(dirname)

    logger.info('Erasing media ...')
    for media_dir in MEDIA_WHITELIST:
        if storage.exists(media_dir):
            delete_files(media_dir)

    logger.info('Done.')



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from django.test import TransactionTestCase
from django.test.utils import override_settings

from kegbot.util import util

from . import models
from .testutils import make_datetime

import copy
from pykeg.backend import get_kegbot_backend


@override_settings(KEGBOT_BACKEND='pykeg.core.testutils.TestBackend')
class StatsTestCase(TransactionTestCase):
    reset_sequences = True

    def setUp(self):
        self.backend = get_kegbot_backend()
        models.User.objects.create_user('guest')

        test_usernames = ('user1', 'user2', 'user3')
        self.users = [self.backend.create_new_user(name, '%s@example.com' % name) for name in test_usernames]

        self.taps = [
            self.backend.create_tap('tap1', 'kegboard.flow0', ticks_per_ml=2.2),
            self.backend.create_tap('tap2', 'kegboard.flow1', ticks_per_ml=2.2),
        ]

        self.keg = self.backend.start_keg('kegboard.flow0', beverage_name='Unknown',
            beverage_type='beer', producer_name='Unknown', style_name='Unknown')

    def testStuff(self):
        site = models.KegbotSite.get()
        stats = site.get_stats()
        self.assertEquals(stats, {})

        now = make_datetime(2012, 1, 2, 12, 00)
        self.maxDiff = None

        d = self.backend.record_drink('kegboard.flow0', ticks=1, volume_ml=100,
            username='user1', pour_time=now)
        expected = util.AttrDict({
            u'volume_by_year': {u'2012': 100.0},
            u'total_pours': 1,
            u'has_guest_pour': False,
            u'greatest_volume_ml': 100.0,
            u'registered_drinkers': [u'user1'],
            u'volume_by_day_of_week': {u'1': 100.0},
            u'greatest_volume_id': d.id,
            u'volume_by_drinker': {u'user1': 100.0},
            u'volume_by_session': {u'1': 100.0},
            u'last_drink_id': d.id,
            u'keg_ids': [d.keg.id],
            u'sessions_count': 1,
            u'average_volume_ml': 100.0,
            u'total_volume_ml': 100.0,
            u'largest_session': {u'session_id': 1, u'volume_ml': 100},
        })
        stats = site.get_stats()
        self.assertDictEqual(expected, stats)

        now = make_datetime(2012, 1, 3, 12, 00)
        d = self.backend.record_drink('kegboard.flow0', ticks=200,
            volume_ml=200, username='user2', pour_time=now)
        stats = site.get_stats()
        expected.total_pours = 2
        expected.greatest_volume_ml = 200.0
        expected.greatest_volume_id = d.id
        expected.registered_drinkers.append(u'user2')
        expected.volume_by_drinker[u'user2'] = 200.0
        expected.last_drink_id = d.id
        expected.average_volume_ml = 150.0
        expected.total_volume_ml = 300.0
        expected.volume_by_day_of_week[u'2'] = 200.0
        expected.volume_by_year[u'2012'] = 300.0
        expected.sessions_count = 2
        expected.volume_by_session = {u'1': 100.0, u'2': 200.0}
        expected.largest_session = {u'session_id': 2, u'volume_ml': 200.0}

        self.assertDictEqual(expected, stats)

        d = self.backend.record_drink('kegboard.flow0', ticks=300,
            volume_ml=300, username='user2', pour_time=now)

        stats = site.get_stats()
        expected.total_pours = 3
        expected.greatest_volume_ml = 300.0
        expected.greatest_volume_id = d.id
        expected.volume_by_drinker[u'user2'] = 500.0
        expected.last_drink_id = d.id
        expected.average_volume_ml = 200.0
        expected.total_volume_ml = 600.0
        expected.volume_by_day_of_week[u'2'] = 500.0
        expected.volume_by_year[u'2012'] = 600.0
        expected.sessions_count = 2
        expected.volume_by_session = {u'1': 100.0, u'2': 500.0}
        expected.largest_session = {u'session_id': 2, u'volume_ml': 500.0}

        self.assertDictEqual(expected, stats)
        previous_stats = copy.copy(stats)

        d = self.backend.record_drink('kegboard.flow0', ticks=300,
            volume_ml=300, pour_time=now)

        stats = site.get_stats()
        self.assertTrue(stats.has_guest_pour)

        self.backend.cancel_drink(d)
        stats = site.get_stats()

        self.assertDictEqual(previous_stats, stats)

    def test_cancel_and_reassign(self):
        drink_data = [
            (100, self.users[0]),
            (200, self.users[1]),
            (200, self.users[2]),
            (500, self.users[0]),
        ]

        drinks = []

        now = make_datetime(2012, 1, 2, 12, 00)
        for volume_ml, user in drink_data:
            d = self.backend.record_drink('kegboard.flow0', ticks=volume_ml,
                username=user.username, volume_ml=volume_ml, pour_time=now)
            drinks.append(d)

        self.assertEquals(600, self.users[0].get_stats().total_volume_ml)
        self.assertEquals(200, self.users[1].get_stats().total_volume_ml)
        self.assertEquals(200, self.users[2].get_stats().total_volume_ml)

        self.assertEquals(1000, models.KegbotSite.get().get_stats().total_volume_ml)

        self.backend.cancel_drink(drinks[0])
        self.assertEquals(500, self.users[0].get_stats().total_volume_ml)
        self.assertEquals(200, self.users[1].get_stats().total_volume_ml)
        self.assertEquals(200, self.users[2].get_stats().total_volume_ml)
        self.assertEquals(900, models.KegbotSite.get().get_stats().total_volume_ml)

        self.backend.assign_drink(drinks[1], self.users[0])
        self.assertEquals(700, self.users[0].get_stats().total_volume_ml)
        self.assertEquals({}, self.users[1].get_stats())
        self.assertEquals(200, self.users[2].get_stats().total_volume_ml)
        self.assertEquals(900, models.KegbotSite.get().get_stats().total_volume_ml)
        self.assertEquals(900, drinks[1].session.get_stats().total_volume_ml)

        # Start a new session.
        now = make_datetime(2013, 1, 2, 12, 00)
        for volume_ml, user in drink_data:
            d = self.backend.record_drink('kegboard.flow0', ticks=volume_ml,
                username=user.username, volume_ml=volume_ml, pour_time=now)
            drinks.append(d)

        self.assertEquals(1300, self.users[0].get_stats().total_volume_ml)
        self.assertEquals(200, self.users[1].get_stats().total_volume_ml)
        self.assertEquals(400, self.users[2].get_stats().total_volume_ml)
        self.assertEquals(1900, models.KegbotSite.get().get_stats().total_volume_ml)
        self.assertEquals(1000, drinks[-1].session.get_stats().total_volume_ml)

        # Delete all stats for some intermediate drinks.
        models.Stats.objects.filter(drink=drinks[-1]).delete()
        models.Stats.objects.filter(drink=drinks[-2]).delete()

        d = self.backend.record_drink('kegboard.flow0', ticks=1111,
            username=user.username, volume_ml=1111, pour_time=now)
        drinks.append(d)

        # Intermediate stats are generated.
        self.assertEquals(3011, models.KegbotSite.get().get_stats().total_volume_ml)
        self.assertEquals(2111, drinks[-1].session.get_stats().total_volume_ml)

    def test_timezone_awareness(self):
        site = models.KegbotSite.get()
        site.timezone = 'US/Pacific'
        site.save()

        drink_data = [
            (100, self.users[0]),
            (200, self.users[1]),
            (200, self.users[2]),
            (500, self.users[0]),
        ]

        drinks = []

        # 1 AM UTC
        now = make_datetime(2012, 1, 2, 1, 0)
        for volume_ml, user in drink_data:
            d = self.backend.record_drink('kegboard.flow0', ticks=volume_ml,
                username=user.username, volume_ml=volume_ml, pour_time=now)
            drinks.append(d)
            self.assertEquals('US/Pacific', d.session.timezone)

        stats = site.get_stats()
        # Assert that stats are recorded for Sunday (day = 0) rather than
        # UTC's monday (day = 1).
        self.assertEquals({'0': 1000.0}, stats.volume_by_day_of_week)
        self.assertEquals(600, self.users[0].get_stats().total_volume_ml)



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from django.conf import settings
from django.core.management.base import NoArgsCommand
from pykeg.backup import backup

import sys


class Command(NoArgsCommand):
    help = u'Erases all data in the current Kegbot system.'

    def handle(self, **options):
        print 'WARNING!'
        print ''
        print '  ************************************************************************'
        print '  This command erases ALL tables and media files in the Kegbot system, and '
        print '  CANNOT BE UNDONE.'
        print ''
        print '    Database: {}'.format(settings.DATABASES['default']['NAME'])
        print '       Media: {}'.format(settings.MEDIA_ROOT)
        print '  ************************************************************************'
        print ''
        print 'Are you SURE you want to continue? '
        print ''

        try:
            response = raw_input('Type ERASE to continue, anything else to abort: ')
        except KeyboardInterrupt:
            response = ''
        print ''
        if response.strip() != 'ERASE':
            print 'Aborted.'
            sys.exit(1)
        backup.erase()



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

import json
import os.path
import random
import datetime
from django.conf import settings
from collections import deque
from uuid import uuid1

from django.contrib.auth.hashers import make_password
from django.core.files import File
from django.core.management.base import BaseCommand
from django.core.management.base import CommandError
from django.db import transaction
from django.utils import timezone

from pykeg.contrib.demomode.views import random_item
from pykeg.backend import get_kegbot_backend
from pykeg.core import defaults
from pykeg.core import models


NUM_SESSIONS = 30
MINUTES_IN_DAY = 60 * 24

MINIMUM_POUR_SIZE_ML = 100
MAXIMUM_POUR_SIZE_ML = 450

# Predictably generate same demo data.
RANDOM_SEED = 42
START_DATE = datetime.datetime(2014, 10, 1)
random.seed(RANDOM_SEED)

"""Generates a bunch of fake drinks from a configuration."""


class LoadDemoDataCommand(BaseCommand):
    args = '<demo_data_dir>'
    help = 'Inserts demo data into the database'

    def handle(self, *args, **options):
        if len(args) < 1:
            raise CommandError('Must specify demo data directory')

        data_dir = os.path.abspath(args[0])
        demo_data = self.load_demo_data(data_dir)

        settings.CELERY_ALWAYS_EAGER = True

        with transaction.atomic():
            if models.User.objects.all().count() or models.Drink.objects.all().count():
                raise CommandError('Database is already loaded.')

            defaults.set_defaults(set_is_setup=True, create_controller=True)

            # Install demo data.
            for username in demo_data['drinkers']:
                user = models.User.objects.create_superuser(username=username, email=None, password=username)
                pw = make_password(username, salt='demo')
                user.password = pw
                user.date_joined = START_DATE
                user.save()

                picture_path = os.path.join(data_dir, 'pictures', 'drinkers', username, 'mugshot.png')
                if os.path.exists(picture_path):
                    p = self.create_picture(picture_path, user=user)
                    user.mugshot = p
                    user.save()

            for beverage in demo_data['beverages']:
                brewer, _ = models.BeverageProducer.objects.get_or_create(name=beverage['brewer_name'])
                beer, _ = models.Beverage.objects.get_or_create(name=beverage['name'], beverage_type='beer',
                    producer=brewer, style=beverage['style'])

                picture_path = beverage.get('image')
                if picture_path:
                    picture = self.create_picture(os.path.join(data_dir, picture_path))
                    beer.picture = picture
                    beer.save()
                else:
                    pass

            sessions = self.build_random_sessions(models.User.objects.all(),
                NUM_SESSIONS, demo_data['shouts'])

            minutes = sessions[-1][0]
            minutes += MINUTES_IN_DAY
            minutes -= (minutes % (MINUTES_IN_DAY))

            session_start = timezone.make_aware(START_DATE, timezone.utc)
            session_start -= datetime.timedelta(hours=session_start.hour + 4,
                minutes=session_start.minute,
                seconds=session_start.second)
            session_start -= datetime.timedelta(minutes=minutes)

            for minute, drinker, volume_ml, shout in sessions:
                date = session_start + datetime.timedelta(minutes=minute)

                picture_path = None
                if random.random() >= 0.25:
                    d = demo_data['pictures'].get(drinker.username)
                    if d:
                        picture_path = d[0]
                        d.rotate()
                self.do_pour(drinker, date, volume_ml, shout, picture_path)

            print 'Demo data loaded.'

    def do_pour(self, user, when, volume_ml, shout, picture_path):
        be = get_kegbot_backend()
        tap = random_item(models.KegTap)

        # End keg if it's near empty.
        if tap.current_keg and tap.current_keg.remaining_volume_ml() < volume_ml:
            be.end_keg(tap.current_keg, when=when)

        # Start keg if the tap is idle.
        if not tap.current_keg:
            beverage = random_item(models.Beverage)
            be.start_keg(tap, beverage=beverage, when=when)

        drink = be.record_drink(tap, ticks=0, volume_ml=volume_ml,
            username=user.username, pour_time=when, shout=shout)

        if picture_path:
            p = self.create_picture(picture_path, user=user,
                session=drink.session, keg=drink.keg)
            drink.picture = p
            drink.save()
        return drink

    def create_picture(self, path, **kwargs):
        f = File(open(path, 'r'))
        p = models.Picture(image=f, **kwargs)
        name = '%s%s' % (uuid1(), os.path.splitext(f.name)[1])
        p.image.save(name, f)
        p.save()
        return p

    def build_random_sessions(self, all_drinkers, count, shouts):
        MIN_POUR_ML = 200
        MAX_POUR_ML = 500

        pours = []
        minute = 0

        for session_number in xrange(count):
            session = []
            num_drinkers = random.randint(1, len(all_drinkers))
            drinkers = random.sample(all_drinkers, num_drinkers)

            # Drink a little more when others are joining
            session_max = 1000 + min(2000, 500 * len(drinkers))

            for drinker in drinkers:
                subminute = random.randint(0, 15)
                session_total = random.randint(500, session_max)
                session_remain = session_total

                while session_remain > 0:
                    if session_remain <= MIN_POUR_ML:
                        drink_volume = session_remain
                    else:
                        drink_volume = random.randint(MIN_POUR_ML, MAX_POUR_ML)

                    shout = ''
                    if shouts and random.random() >= 0.60:
                        shout = shouts[0]
                        shouts.rotate(1)

                    session.append((minute + subminute, drinker, drink_volume, shout))
                    subminute += int(drink_volume / 40) + random.randint(0, 30)  # 40 mL/minute min plus noise
                    session_remain -= drink_volume

            # session.sort()
            # for minute, drinker, volume, shout in session:
            #     print '%08s %016s %04s  %s' % (minute, drinker, volume, shout)
            # print ''

            pours += session

            # wait 1 - 3 days for next session
            minute += random.randint(1, 3) * MINUTES_IN_DAY
            minute -= (minute % (MINUTES_IN_DAY))
            minute += random.randint(-120, 120)

        pours.sort()
        return pours

    def load_demo_data(self, path):
        demo_data_file = os.path.join(path, 'demo_data.json')
        if not os.path.isfile(demo_data_file):
            raise CommandError('Missing demo_data file: %s' % demo_data_file)

        demo_data = json.loads(open(demo_data_file).read())
        demo_data['pictures'] = {}

        for drinker_name in demo_data['drinkers']:
            demo_data['pictures'][drinker_name] = deque()

            pictures_dir = os.path.join(path, 'pictures', 'drinkers', drinker_name)
            if not os.path.isdir(pictures_dir):
                continue

            for picture_name in (p for p in os.listdir(pictures_dir) if not p.startswith('.')):
                if picture_name == 'mugshot.png':
                    continue
                picture_path = os.path.join(pictures_dir, picture_name)
                demo_data['pictures'][drinker_name].append(picture_path)
            random.shuffle(demo_data['pictures'][drinker_name])

        demo_data['shouts'] = deque(demo_data['shouts'])
        random.shuffle(demo_data['shouts'])

        return demo_data

Command = LoadDemoDataCommand



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

import os

from django.core.files.storage import get_storage_class
from django.core.management.base import BaseCommand
from pykeg.backup import backup
from optparse import make_option
from pykeg.backend.backends import UnknownBaseUrlException


class Command(BaseCommand):
    help = u'Creates a zipfile backup of the current Kegbot system.'
    option_list = BaseCommand.option_list + (
        make_option('--no_media', action='store_true', dest='no_media', default=False,
            help='Skip media during backup.'),
    )

    def handle(self, **options):
        location = backup.backup(include_media=not options.get('no_media'))
        storage = get_storage_class()()

        path = location
        if hasattr(storage, 'location'):
            path = os.path.join(storage.location, path)
        print 'Backup complete!'
        print 'Path: {}'.format(path)
        try:
            print 'URL: {}'.format(storage.url(location))
        except (NotImplementedError, UnknownBaseUrlException):
            pass



#!/usr/bin/env python
#
# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
from pykeg.core.kb_common import LOCAL_SETTINGS_SEARCH_DIRS

_DEBUG = False

_CUSTOM_SEARCH_DIR = os.getenv('KEGBOT_SETTINGS_DIR')
if _CUSTOM_SEARCH_DIR:
    SEARCH_DIRS = [os.path.expanduser(_CUSTOM_SEARCH_DIR)]
else:
    SEARCH_DIRS = [os.path.expanduser(d) for d in LOCAL_SETTINGS_SEARCH_DIRS]


def _Debug(message):
    if _DEBUG:
        sys.stderr.write('importhacks: %s\n' % (message,))


def _Warning(message):
    if _DEBUG:
        sys.stderr.write('importhacks: %s\n' % (message,))


def _AddToSysPath(paths):
    for path in paths:
        path = os.path.abspath(path)
        if path not in sys.path:
            _Debug('Adding to sys.path: %s' % path)
            sys.path.append(path)
        else:
            _Debug('Already in sys.path: %s' % path)


def _ExtendSysPath():
    """ Add some paths where we'll look for user settings. """
    paths = []
    for dir in SEARCH_DIRS:
        paths.append(dir)
        if _DEBUG:
            test_settings = os.path.join(dir, 'local_settings.py')
            if os.path.exists(test_settings):
                _Debug('%s exists' % test_settings)
            else:
                _Debug('%s does NOT exist' % test_settings)

    _AddToSysPath(paths)


def _SetDjangoSettingsEnv(settings='pykeg.settings'):
    """ Set django settings if not set. """
    if not os.environ.get('DJANGO_SETTINGS_MODULE'):
        _Debug('Setting DJANGO_SETTINGS_MODULE=%s' % (settings,))
        os.environ['DJANGO_SETTINGS_MODULE'] = settings


def _FixAll():
    try:
        import pykeg
        pykeg_dir = os.path.dirname(pykeg.__file__)
        _Debug('Pykeg loaded from dir: %s' % pykeg_dir)
    except ImportError:
        _Warning('Error: pykeg could not be imported')
        sys.exit(1)

    _ExtendSysPath()
    _SetDjangoSettingsEnv()

    try:
        import local_settings
        common_dir = os.path.dirname(local_settings.__file__)
        _Debug('local_settings loaded from dir: %s' % common_dir)
    except ImportError:
        _Warning('Warning: local_settings could not be imported')

if __name__ == '__main__' or os.environ.get('DEBUG_IMPORT_HACKS'):
    # When run as a program, or DEBUG_IMPORT_HACKS is set: print debug info
    _DEBUG = True

_FixAll()



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Unittests for email notification backend ."""

from django.core import mail
from django.test import TestCase
from django.test.utils import override_settings
from pykeg.backend import get_kegbot_backend
from pykeg.core import models
from pykeg.core import defaults
from pykeg.core import kb_common


@override_settings(KEGBOT_BACKEND='pykeg.core.testutils.TestBackend')
@override_settings(EMAIL_BACKEND='django.core.mail.backends.locmem.EmailBackend')
@override_settings(EMAIL_FROM_ADDRESS='test-from@example')
class EmailNotificationBackendTestCase(TestCase):
    def setUp(self):
        self.backend = get_kegbot_backend()
        defaults.set_defaults(set_is_setup=True, create_controller=True)

        self.user = models.User.objects.create(username='notification_user',
            email='test@example')

        self.prefs = models.NotificationSettings.objects.create(user=self.user,
            backend='pykeg.notification.backends.email.EmailNotificationBackend',
            keg_tapped=False, session_started=False, keg_volume_low=False,
            keg_ended=False)

    def test_keg_tapped(self):
        self.prefs.keg_tapped = True
        self.prefs.save()
        self.assertEquals(0, len(mail.outbox))

        keg = self.backend.start_keg(defaults.METER_NAME_0, beverage_name='Unknown',
            beverage_type='beer', producer_name='Unknown', style_name='Unknown')
        self.assertEquals(1, len(mail.outbox))

        msg = mail.outbox[0]
        self.assertEquals('[My Kegbot] New keg tapped: Keg %s: Unknown by Unknown' % keg.id,
            msg.subject)
        self.assertEquals(['test@example'], msg.to)
        self.assertEquals('test-from@example', msg.from_email)

        expected_body_plain = '''A new keg of Unknown by Unknown was just tapped on My Kegbot!

Track it here: http://localhost:1234/kegs/%s

You are receiving this e-mail because you have notifications enabled
on My Kegbot.  To change your settings, visit http://localhost:1234/account.''' % (keg.id,)

        self.assertMultiLineEqual(expected_body_plain, msg.body)

        expected_body_html = '''<p>
A new keg of <b>Unknown by Unknown</b> was just tapped on
<b><a href="http://localhost:1234">My Kegbot</a></b>!
</p>

<p>
Track it <a href="http://localhost:1234/kegs/%s">here</a>.
</p>

<p>
You are receiving this e-mail because you have notifications enabled
on <a href="http://localhost:1234">My Kegbot</a>.  To change your settings, visit
http://localhost:1234/account.
</p>''' % (keg.id,)

        self.assertEquals(1, len(msg.alternatives))
        self.assertMultiLineEqual(expected_body_html, msg.alternatives[0][0])
        self.assertEquals('text/html', msg.alternatives[0][1])

    def test_session_started(self):
        self.prefs.session_started = True
        self.prefs.save()
        self.assertEquals(0, len(mail.outbox))

        self.backend.start_keg(defaults.METER_NAME_0, beverage_name='Unknown',
            beverage_type='beer', producer_name='Unknown', style_name='Unknown')
        drink = self.backend.record_drink(defaults.METER_NAME_0, ticks=500)
        self.assertEquals(1, len(mail.outbox))

        msg = mail.outbox[0]
        self.assertEquals('[My Kegbot] A new session (session %s) has started.' % (
            drink.session.id,), msg.subject)
        self.assertEquals(['test@example'], msg.to)
        self.assertEquals('test-from@example', msg.from_email)

        expected_body_plain = '''A new session was just kicked off on My Kegbot.

You can follow the session here: http://localhost:1234/s/%s

You are receiving this e-mail because you have notifications enabled
on My Kegbot.  To change your settings, visit http://localhost:1234/account.''' % (drink.session.id,)

        self.assertMultiLineEqual(expected_body_plain, msg.body)

        expected_body_html = '''<p>
A new session was just kicked off on <a href="http://localhost:1234">My Kegbot</a>.
</p>

<p>
You can follow the session <a href="http://localhost:1234/s/%s">here</a>.
</p>

<p>
You are receiving this e-mail because you have notifications enabled
on <a href="http://localhost:1234">My Kegbot</a>.  To change your settings, visit
http://localhost:1234/account.
</p>''' % (drink.session.id,)

        self.assertEquals(1, len(msg.alternatives))
        self.assertMultiLineEqual(expected_body_html, msg.alternatives[0][0])
        self.assertEquals('text/html', msg.alternatives[0][1])

    def test_keg_volume_low(self):
        self.prefs.keg_volume_low = True
        self.prefs.save()
        self.assertEquals(0, len(mail.outbox))

        keg = self.backend.start_keg(defaults.METER_NAME_0, beverage_name='Unknown',
            beverage_type='beer', producer_name='Unknown', style_name='Unknown')
        self.backend.record_drink(defaults.METER_NAME_0, ticks=500,
            volume_ml=keg.full_volume_ml * (1 - kb_common.KEG_VOLUME_LOW_PERCENT))
        self.assertEquals(1, len(mail.outbox))

        msg = mail.outbox[0]
        self.assertEquals('[My Kegbot] Volume low on keg %s (Unknown by Unknown)' % keg.id,
            msg.subject)
        self.assertEquals(['test@example'], msg.to)
        self.assertEquals('test-from@example', msg.from_email)

        expected_body_plain = '''Keg %s (Unknown by Unknown) is 15.0%% full.

See full statistics here: http://localhost:1234/kegs/%s

You are receiving this e-mail because you have notifications enabled
on My Kegbot.  To change your settings, visit http://localhost:1234/account.''' % (keg.id, keg.id)

        self.assertMultiLineEqual(expected_body_plain, msg.body)

        expected_body_html = '''<p>
Keg %s (Unknown by Unknown) is <b>15.0</b>%% full.
</p>

<p>
See full statistics <a href="http://localhost:1234/kegs/%s">here</a>.
</p>

<p>
You are receiving this e-mail because you have notifications enabled
on <a href="http://localhost:1234">My Kegbot</a>.  To change your settings, visit
http://localhost:1234/account.
</p>''' % (keg.id, keg.id)

        self.assertEquals(1, len(msg.alternatives))
        self.assertMultiLineEqual(expected_body_html, msg.alternatives[0][0])
        self.assertEquals('text/html', msg.alternatives[0][1])

    def test_keg_ended(self):
        self.prefs.keg_ended = True
        self.prefs.save()
        self.assertEquals(0, len(mail.outbox))

        keg = self.backend.start_keg(defaults.METER_NAME_0, beverage_name='Unknown',
            beverage_type='beer', producer_name='Unknown', style_name='Unknown')
        self.assertEquals(0, len(mail.outbox))
        self.backend.end_keg(keg)
        self.assertEquals(1, len(mail.outbox))

        msg = mail.outbox[0]
        self.assertEquals('[My Kegbot] Keg ended: Keg %s: Unknown by Unknown' % keg.id, msg.subject)
        self.assertEquals(['test@example'], msg.to)
        self.assertEquals('test-from@example', msg.from_email)

        expected_body_plain = '''Keg %s of Unknown by Unknown was just finished on My Kegbot.

See final statistics here: http://localhost:1234/kegs/%s

You are receiving this e-mail because you have notifications enabled
on My Kegbot.  To change your settings, visit http://localhost:1234/account.''' % (keg.id, keg.id)

        self.assertMultiLineEqual(expected_body_plain, msg.body)

        expected_body_html = '''<p>
Keg %s (Unknown by Unknown) was just finished on
<b><a href="http://localhost:1234">My Kegbot</a></b>.
</p>

<p>
See final statistics <a href="http://localhost:1234/kegs/%s">here</a>.
</p>

<p>
You are receiving this e-mail because you have notifications enabled
on <a href="http://localhost:1234">My Kegbot</a>.  To change your settings, visit
http://localhost:1234/account.
</p>''' % (keg.id, keg.id)

        self.assertEquals(1, len(msg.alternatives))
        self.assertMultiLineEqual(expected_body_html, msg.alternatives[0][0])
        self.assertEquals('text/html', msg.alternatives[0][1])



from .exceptions import BackupError


def engine_name():
    return 'unknown'


def is_installed(*args, **kwargs):
    raise BackupError('Engine unsupported.')


def dump(*args, **kwargs):
    raise BackupError('Engine unsupported.')


def restore(*args, **kwargs):
    raise BackupError('Engine unsupported.')


def erase(*args, **kwargs):
    raise BackupError('Engine unsupported.')



from django import forms
from django.conf import settings
from django.contrib.humanize.templatetags.humanize import naturaltime
from crispy_forms.helper import FormHelper
from crispy_forms.layout import Layout, Submit, Field, HTML
from crispy_forms.bootstrap import FormActions

from kegbot.util import units
from pykeg.backend import get_kegbot_backend
from pykeg.core import keg_sizes
from pykeg.core import models

ALL_TAPS = models.KegTap.objects.all()
ALL_KEGS = models.Keg.objects.all()
ALL_METERS = models.FlowMeter.objects.all()
ALL_TOGGLES = models.FlowToggle.objects.all()
ALL_THERMOS = models.ThermoSensor.objects.all()


class ChangeKegForm(forms.Form):
    keg_size = forms.ChoiceField(choices=keg_sizes.CHOICES,
        initial=keg_sizes.HALF_BARREL,
        required=True)

    initial_volume = forms.FloatField(label='Initial Volume', initial=0.0,
        required=False, help_text='Keg\'s Initial Volume')

    beer_name = forms.CharField(required=False)  # legacy
    brewer_name = forms.CharField(required=False)  # legacy

    beverage_name = forms.CharField(label='Beer Name', required=False)
    beverage_id = forms.CharField(widget=forms.HiddenInput(), required=False)
    producer_name = forms.CharField(label='Brewer', required=False)
    producer_id = forms.CharField(widget=forms.HiddenInput(), required=False)
    style_name = forms.CharField(required=True, label='Style',
      help_text='Example: Pale Ale, Stout, etc.')

    helper = FormHelper()
    helper.form_class = 'form-horizontal beer-select'
    helper.layout = Layout(
        Field('beverage_name', css_class='input-xlarge'),
        Field('beverage_id', type='hidden'),
        Field('producer_name', css_class='input-xlarge'),
        Field('producer_id', type='hidden'),
        Field('style_name', css_class='input-xlarge'),
        Field('keg_size', css_class='input-xlarge'),
        Field('initial_volume', css_class='input-volume'),
        FormActions(
            Submit('submit_change_keg_form', 'Activate Keg', css_class='btn-primary'),
        )
    )

    def clean_beverage_name(self):
        beverage_name = self.cleaned_data.get('beverage_name')
        if not beverage_name:
            beverage_name = self.cleaned_data.get('beer_name')
            if not beverage_name:
                raise forms.ValidationError('Must specify a beverage name')
            self.cleaned_data['beverage_name'] = beverage_name
        return beverage_name

    def clean_producer_name(self):
        producer_name = self.cleaned_data.get('producer_name')
        if not producer_name:
            producer_name = self.cleaned_data.get('brewer_name')
            if not producer_name:
                raise forms.ValidationError('Must specify a producer name')
            self.cleaned_data['producer_name'] = producer_name
        return producer_name

    def save(self, tap):
        if not self.is_valid():
            raise ValueError('Form is not valid.')
        b = get_kegbot_backend()

        if tap.is_active():
            b.end_keg(tap.current_keg)

        keg_size = self.cleaned_data.get('keg_size')
        full_volume_ml = self.cleaned_data.get('full_volume_ml')

        if keg_size != 'other':
            full_volume_ml = None
        else:
            full_volume_ml = self.cleaned_data.get('initial_volume')

        # TODO(mikey): Support non-beer beverage types.
        cd = self.cleaned_data
        keg = b.start_keg(tap, beverage_name=cd['beverage_name'], producer_name=cd['producer_name'],
            beverage_type='beer', style_name=cd['style_name'], keg_type=cd['keg_size'],
            full_volume_ml=full_volume_ml)

        if cd.get('description'):
            keg.description = cd['description']
            keg.save()


class EndKegForm(forms.Form):
    keg = forms.ModelChoiceField(queryset=ALL_KEGS, widget=forms.HiddenInput)

    helper = FormHelper()
    helper.form_class = 'form-horizontal beer-select'
    helper.layout = Layout(
        Field('keg', type='hidden'),
        FormActions(
            Submit('submit_end_keg_form', 'End Keg', css_class='btn-danger'),
        )
    )


class TapForm(forms.ModelForm):
    class FlowMeterModelChoiceField(forms.ModelChoiceField):
        def label_from_instance(self, meter):
            if meter.tap:
                return u'{} (connected to {})'.format(meter, meter.tap.name)
            else:
                return unicode(meter)

    class FlowToggleModelChoiceField(forms.ModelChoiceField):
        def label_from_instance(self, toggle):
            if toggle.tap:
                return u'{} (connected to {})'.format(toggle, toggle.tap.name)
            else:
                return unicode(toggle)

    class ThermoSensorModelChoiceField(forms.ModelChoiceField):
        def label_from_instance(self, sensor):
            last_log = sensor.LastLog()
            if last_log:
                return u'{} (Last report: {})'.format(sensor, naturaltime(last_log.time))
            else:
                return unicode(sensor)

    meter = FlowMeterModelChoiceField(queryset=ALL_METERS, required=False,
        empty_label='Not connected.',
        help_text='Tap is routed thorough this flow meter. If unset, reporting is disabled.')

    toggle = FlowToggleModelChoiceField(queryset=ALL_TOGGLES, required=False,
        empty_label='Not connected.',
        help_text='Optional flow toggle (usually a relay/valve) connected to this tap.')

    temperature_sensor = ThermoSensorModelChoiceField(queryset=ALL_THERMOS, required=False,
        empty_label='No sensor.',
        help_text='Optional sensor monitoring the temperature at this tap.')

    class Meta:
        model = models.KegTap
        fields = ('name', 'notes', 'temperature_sensor', 'sort_order')

    def __init__(self, *args, **kwargs):
        super(TapForm, self).__init__(*args, **kwargs)

        if self.instance:
            self.fields['meter'].initial = self.instance.current_meter()
            self.fields['toggle'].initial = self.instance.current_toggle()
            self.fields['temperature_sensor'].initial = self.instance.temperature_sensor

    def save(self, commit=True):
        if not commit:
            raise ValueError('TapForm does not support commit=False')
        instance = super(TapForm, self).save(commit=True)
        b = get_kegbot_backend()
        b.connect_meter(instance, self.cleaned_data['meter'])
        b.connect_toggle(instance, self.cleaned_data['toggle'])
        return instance

    helper = FormHelper()
    helper.form_class = 'form-horizontal'
    helper.layout = Layout(
        Field('name', css_class='input-xlarge'),
        Field('meter', css_class='input-xlarge'),
        Field('toggle', css_class='input-xlarge'),
        Field('temperature_sensor', css_class='input-xlarge'),
        Field('sort_order', css_class='input-xlarge'),
        Field('notes', css_class='input-block-level', rows='3'),
        FormActions(
            Submit('submit_tap_form', 'Save Settings', css_class='btn-success'),
        )
    )


class DeleteTapForm(forms.Form):
    helper = FormHelper()
    helper.form_class = 'form-horizontal'
    helper.layout = Layout(
        FormActions(
            Submit('submit_delete_tap_form', 'Delete Tap', css_class='btn-danger'),
        )
    )


class KegForm(forms.Form):
    keg_size = forms.ChoiceField(choices=keg_sizes.CHOICES,
        initial=keg_sizes.HALF_BARREL,
        required=True)

    initial_volume = forms.FloatField(label='Initial Volume', initial=0.0,
        required=False, help_text='Keg\'s Initial Volume')

    beer_name = forms.CharField(required=False)  # legacy
    brewer_name = forms.CharField(required=False)  # legacy

    beverage_name = forms.CharField(label='Beer Name', required=False)
    beverage_id = forms.CharField(widget=forms.HiddenInput(), required=False)
    producer_name = forms.CharField(label='Brewer', required=False)
    producer_id = forms.CharField(widget=forms.HiddenInput(), required=False)
    style_name = forms.CharField(required=True, label='Style',
      help_text='Example: Pale Ale, Stout, etc.')

    description = forms.CharField(max_length=256, label='Description',
        widget=forms.Textarea(), required=False,
        help_text='Optional user-visible description of the keg.')
    notes = forms.CharField(label='Notes', required=False, widget=forms.Textarea(),
        help_text='Optional private notes about this keg, viewable only by admins.')
    connect_to = forms.ModelChoiceField(queryset=ALL_TAPS, label='Connect To',
        required=False,
        help_text='If selected, immediately activates the keg on this tap. '
            '(Any existing keg will be ended.)')

    helper = FormHelper()
    helper.form_class = 'form-horizontal beer-select'
    helper.layout = Layout(
        Field('beverage_name', css_class='input-xlarge'),
        Field('beverage_id', type='hidden'),
        Field('producer_name', css_class='input-xlarge'),
        Field('producer_id', type='hidden'),
        Field('style_name', css_class='input-xlarge'),
        Field('keg_size', css_class='input-xlarge'),
        Field('initial_volume', css_class='input-volume'),
        Field('description', css_class='input-block-level', rows='3'),
        Field('notes', css_class='input-block-level', rows='3'),
        Field('connect_to', css_class='input-block-level'),
        FormActions(
            Submit('submit_add_keg', 'Save', css_class='btn-primary'),
        )
    )

    def clean_beverage_name(self):
        beverage_name = self.cleaned_data.get('beverage_name')
        if not beverage_name:
            beverage_name = self.cleaned_data.get('beer_name')
            if not beverage_name:
                raise forms.ValidationError('Must specify a beverage name')
            self.cleaned_data['beverage_name'] = beverage_name
        return beverage_name

    def clean_producer_name(self):
        producer_name = self.cleaned_data.get('producer_name')
        if not producer_name:
            producer_name = self.cleaned_data.get('brewer_name')
            if not producer_name:
                raise forms.ValidationError('Must specify a producer name')
            self.cleaned_data['producer_name'] = producer_name
        return producer_name

    def save(self):
        if not self.is_valid():
            raise ValueError('Form is not valid.')
        keg_size = self.cleaned_data.get('keg_size')
        if keg_size != 'other':
            full_volume_ml = None
        else:
            full_volume_ml = self.cleaned_data.get('initial_volume')

        # TODO(mikey): Support non-beer beverage types.
        cd = self.cleaned_data
        b = get_kegbot_backend()
        keg = b.create_keg(beverage_name=cd['beverage_name'], producer_name=cd['producer_name'],
            beverage_type='beer', style_name=cd['style_name'], keg_type=cd['keg_size'],
            full_volume_ml=full_volume_ml, notes=cd['notes'], description=cd['description'])

        tap = cd['connect_to']
        if tap:
            if tap.is_active():
                b.end_keg(tap.current_keg)
            b.attach_keg(tap, keg)

        return keg


class EditKegForm(forms.ModelForm):
    class Meta:
        model = models.Keg
        fields = ('type', 'keg_type', 'full_volume_ml', 'spilled_ml', 'description', 'notes',)
        labels = {
            'full_volume_ml': ('Full/Initial Volume'),
            'spilled_ml': ('Spilled Volume'),
        }

    helper = FormHelper()
    helper.form_class = 'form-horizontal'
    helper.layout = Layout(
        Field('type', css_class='input-block-level', rows='3'),
        Field('keg_type', css_class='input-block-level', rows='3'),
        Field('full_volume_ml', css_class='input-volume'),
        Field('spilled_ml', css_class='input-volume'),
        Field('description', css_class='input-block-level', rows='3'),
        Field('notes', css_class='input-block-level', rows='3'),
        FormActions(
            Submit('submit_edit_keg', 'Save Keg', css_class='btn-primary'),
            HTML("""<a href="#cancelModal" role="button" class="btn btn-danger" data-toggle="modal">
                    <i class="icon-trash icon-white"></i> Permanently Delete</a>""")
        )
    )


class GeneralSiteSettingsForm(forms.ModelForm):
    class Meta:
        model = models.KegbotSite
        fields = (
            'title',
            'enable_sensing',
            'enable_users',
            'privacy',
            'registration_mode',
        )

    helper = FormHelper()
    helper.form_class = 'form-horizontal'
    helper.layout = Layout(
        Field('title', css_class='input-xlarge'),
        Field('enable_sensing', css_class='input-xlarge'),
        Field('enable_users', css_class='input-xlarge'),
        Field('privacy', css_class='input-xlarge'),
        Field('registration_mode', css_class='input-xlarge'),
        FormActions(
            Submit('submit', 'Save Settings', css_class='btn-primary'),
        )
    )


class LocationSiteSettingsForm(forms.ModelForm):
    guest_image = forms.ImageField(required=False,
        help_text='Custom image for the "guest" user.')

    class Meta:
        model = models.KegbotSite
        fields = (
            'volume_display_units',
            'temperature_display_units',
            'timezone',
        )

    helper = FormHelper()
    helper.form_class = 'form-horizontal'
    helper.layout = Layout(
        Field('volume_display_units', css_class='input-xlarge'),
        Field('temperature_display_units', css_class='input-xlarge'),
        Field('timezone'),
        FormActions(
            Submit('submit', 'Save Settings', css_class='btn-primary'),
        )
    )


class AdvancedSiteSettingsForm(forms.ModelForm):
    class Meta:
        model = models.KegbotSite
        fields = (
            'session_timeout_minutes',
            'google_analytics_id',
            'check_for_updates',
        )

    helper = FormHelper()
    helper.form_class = 'form-horizontal'
    helper.layout = Layout(
        Field('session_timeout_minutes'),
        Field('google_analytics_id'),
        Field('check_for_updates'),
        FormActions(
            Submit('submit', 'Save Settings', css_class='btn-primary'),
        )
    )


class BeverageForm(forms.ModelForm):
    class Meta:
        model = models.Beverage
        fields = ('name', 'style', 'producer', 'vintage_year', 'abv_percent',
            'original_gravity', 'specific_gravity', 'ibu', 'srm', 'color_hex',
            'star_rating', 'untappd_beer_id', 'description')

    new_image = forms.ImageField(required=False,
        help_text='Set/replace image for this beer type.')

    helper = FormHelper()
    helper.form_class = 'form-horizontal'
    helper.layout = Layout(
        Field('name', css_class='input-xlarge'),
        Field('style', css_class='input-xlarge'),
        Field('producer'),
        Field('vintage_year'),
        Field('abv_percent'),
        Field('original_gravity'),
        Field('specific_gravity'),
        Field('ibu'),
        Field('srm'),
        Field('color_hex'),
        Field('star_rating'),
        Field('untappd_beer_id'),
        Field('description'),
        Field('new_image'),
        FormActions(
            Submit('submit', 'Save', css_class='btn-primary'),
        )
    )


class BeverageProducerForm(forms.ModelForm):
    class Meta:
        model = models.BeverageProducer

    helper = FormHelper()
    helper.form_class = 'form-horizontal'
    helper.layout = Layout(
        Field('name', css_class='input-xlarge'),
        Field('country', css_class='input-xlarge'),
        Field('origin_state', css_class='input-xlarge'),
        Field('origin_city', css_class='input-xlarge'),
        Field('is_homebrew'),
        Field('url'),
        Field('description'),
        FormActions(
            Submit('submit', 'Save', css_class='btn-primary'),
        )
    )


class FindUserForm(forms.Form):
    username = forms.CharField()


class UserForm(forms.ModelForm):
    class Meta:
        model = models.User
        fields = (
            'username',
            'email',
            'password',
            'is_staff',
        )

    helper = FormHelper()
    helper.form_class = 'form-horizontal'
    helper.layout = Layout(
        Field('username', css_class='input-xlarge'),
        Field('email', css_class='input-xlarge'),
        'password',
        'is_staff',
        FormActions(
            Submit('submit', 'Save', css_class='btn-primary'),
        )
    )


class UserProfileForm(forms.ModelForm):
    class Meta:
        model = models.User
        fields = ('display_name',)
        if settings.EMBEDDED:
            fields = ('email', ) + fields

    new_mugshot = forms.ImageField(required=False)

    def save(self, *args, **kwargs):
        user = super(UserProfileForm, self).save(*args, **kwargs)
        image = self.cleaned_data.get('new_mugshot')
        if image:
            pic = models.Picture.objects.create(user=user)
            pic.image.save(image.name, image)
            pic.save()
            user.mugshot = pic
            user.save()
        return user


class TokenForm(forms.ModelForm):
    class Meta:
        model = models.AuthenticationToken
        fields = (
            'nice_name',
            'enabled',
        )
    username = forms.CharField(required=False)

    helper = FormHelper()
    helper.form_class = 'form-horizontal user-select'
    helper.layout = Layout(
        Field('username', css_class='input-xlarge'),
        Field('nice_name', css_class='input-xlarge'),
        'enabled',
        FormActions(
            Submit('submit', 'Save', css_class='btn-primary'),
        )
    )

    def clean_username(self):
        username = self.cleaned_data['username']
        if username == '':
            self.cleaned_data['user'] = None
            return
        try:
            self.cleaned_data['user'] = models.User.objects.get(username=username)
        except models.User.DoesNotExist:
            raise forms.ValidationError('Invalid username; use a complete user name or leave blank.')
        return username


class AddTokenForm(forms.ModelForm):
    class Meta:
        model = models.AuthenticationToken
        fields = (
            'auth_device',
            'token_value',
            'enabled',
        )
    CHOICES = (
        ('core.rfid', 'RFID'),
        ('core.onewire', 'OneWire/iButton'),
        ('nfc', 'NFC'),
    )
    auth_device = forms.ChoiceField(choices=CHOICES)
    username = forms.CharField(required=False)

    helper = FormHelper()
    helper.form_class = 'form-horizontal user-select'
    helper.layout = Layout(
        Field('auth_device', css_class='input-xlarge'),
        Field('token_value', css_class='input-xlarge'),
        Field('username', css_class='input-xlarge'),
        'enabled',
        FormActions(
            Submit('submit', 'Save', css_class='btn-primary'),
        )
    )

    def clean_username(self):
        username = self.cleaned_data['username']
        if username == '':
            self.cleaned_data['user'] = None
            return
        try:
            self.cleaned_data['user'] = models.User.objects.get(username=username)
        except models.User.DoesNotExist:
            raise forms.ValidationError('Invalid username; use a complete user name or leave blank.')
        return username


class CancelDrinkForm(forms.Form):
    pass


class ReassignDrinkForm(forms.Form):
    username = forms.CharField(required=True)


class ChangeDrinkVolumeForm(forms.Form):
    UNIT_CHOICES = (
        ('mL', 'mL'),
        ('oz', 'oz')
    )
    units = forms.ChoiceField(required=True, choices=UNIT_CHOICES)
    volume = forms.FloatField(required=True, min_value=0)

    def clean_volume(self):
        volume = self.cleaned_data['volume']
        if self.cleaned_data['units'] == 'oz':
            self.cleaned_data['volume_ml'] = float(units.Quantity(volume, units.UNITS.Ounce).InMilliliters())
        else:
            self.cleaned_data['volume_ml'] = volume
        return volume


class RecordDrinkForm(forms.Form):
    units = forms.ChoiceField(required=True, choices=ChangeDrinkVolumeForm.UNIT_CHOICES)
    volume = forms.FloatField(required=True, min_value=0)
    username = forms.CharField(required=False)

    def clean_username(self):
        username = self.cleaned_data['username']
        if username == '':
            self.cleaned_data['user'] = None
            return
        try:
            self.cleaned_data['user'] = models.User.objects.get(username=username)
        except models.User.DoesNotExist:
            raise forms.ValidationError('Invalid username; use a complete user name or leave blank.')
        return username

    def clean_volume(self):
        volume = self.cleaned_data['volume']
        if self.cleaned_data['units'] == 'oz':
            self.cleaned_data['volume_ml'] = float(units.Quantity(volume, units.UNITS.Ounce).InMilliliters())
        else:
            self.cleaned_data['volume_ml'] = volume
        return volume


class TestEmailForm(forms.Form):
    address = forms.CharField(required=False)

    def clean_address(self):
        address = self.cleaned_data['address']
        if address == '':
            self.cleaned_data['address'] = None
            return
        return address


class NewFlowMeterForm(forms.ModelForm):
    class Meta:
        model = models.FlowMeter
        fields = ('port_name', 'ticks_per_ml', 'controller')


class UpdateFlowMeterForm(forms.ModelForm):
    class Meta:
        model = models.FlowMeter
        fields = ('ticks_per_ml',)


class AddFlowMeterForm(forms.ModelForm):
    class Meta:
        model = models.FlowMeter
        fields = ('port_name', 'ticks_per_ml', 'controller')

    helper = FormHelper()
    helper.form_class = 'form-horizontal'
    helper.layout = Layout(
        Field('port_name'),
        Field('ticks_per_ml'),
        Field('controller', type='hidden'),
        FormActions(
            Submit('add_flow_meter', 'Add Flow Meter', css_class='btn-primary'),
        )
    )


class FlowToggleForm(forms.ModelForm):
    class Meta:
        model = models.FlowToggle
        fields = ('port_name', 'controller')


class AddFlowToggleForm(forms.ModelForm):
    class Meta:
        model = models.FlowToggle
        fields = ('port_name', 'controller')

    helper = FormHelper()
    helper.form_class = 'form-horizontal'
    helper.layout = Layout(
        Field('port_name'),
        Field('controller', type='hidden'),
        FormActions(
            Submit('add_flow_toggle', 'Add Flow Toggle', css_class='btn-primary'),
        )
    )


class DeleteControllerForm(forms.Form):
    helper = FormHelper()
    helper.form_class = 'form-horizontal user-select'
    helper.layout = Layout(
        FormActions(
            Submit('delete_controller', 'Delete Controller', css_class='btn-danger'),
        )
    )


class ControllerForm(forms.ModelForm):
    class Meta:
        model = models.Controller
        fields = ('name', 'model_name', 'serial_number')


class LinkDeviceForm(forms.Form):
    code = forms.CharField(required=True,
        help_text='Link code shown on device.')
    helper = FormHelper()
    helper.form_class = 'form-horizontal'
    helper.layout = Layout(
        'code',
        FormActions(
            Submit('link_device', 'Link Device', css_class='btn-success'),
        )
    )



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

import datetime

from django.utils import timezone
from kegbot.util import units
from kegbot.util import util

from pykeg.core import models


class ChartError(Exception):
    """Base chart exception."""


def format_volume(volume_ml, chart_kwargs):
    metric_volumes = chart_kwargs.get('metric_volumes', False)
    if metric_volumes:
        return volume_ml / 1000.0, 'L'
    else:
        return units.Quantity(volume_ml).InPints(), 'pints'


def format_temperature(temp_c, chart_kwargs):
    use_c = chart_kwargs.get('temperature_units', None) == 'c'
    if use_c:
        return temp_c
    else:
        return util.CtoF(temp_c)


def chart_temp_sensor(sensor, *args, **kwargs):
    """ Shows a simple line plot of a specific temperature sensor.

    Syntax:
      {% chart temp_sensor <sensorname> width height %}
    Args:
      sensorname - the nice_name of a ThermoSensor
    """
    if not isinstance(sensor, models.ThermoSensor):
        raise ChartError('Bad sensor given')

    hours = 6
    now = timezone.now()
    start = now - (datetime.timedelta(hours=hours))
    start = start - (datetime.timedelta(seconds=start.second))

    points = sensor.thermolog_set.filter(time__gte=start).order_by('time')

    curr = start
    temps = []
    have_temps = False
    for point in points:
        temp = format_temperature(point.temp, kwargs)
        while curr <= point.time:
            curr += datetime.timedelta(minutes=1)
            if curr < point.time:
                temps.append(None)
            else:
                temps.append(temp)
                have_temps = True

    if not have_temps:
        raise ChartError('Not enough data')

    res = {
        'series': [
          {
              'data': temps,
              'marker': {
                  'enabled': False,
              },
          },
        ],
        'tooltip': {
            'enabled': True,
        },
        'xAxis': {
            'categories': ['Temperature'],
            'labels': {
                'enabled': False,
            },
            'tickInterval': 60,
        },
        'yAxis': {
            'labels': {
                'enabled': True,
            },
            'tickInterval': 1,
        },
    }
    return res


def chart_volume_by_weekday(stats, *args, **kwargs):
    """ Shows a histogram of volume by day of the week.

    Syntax:
      {% chart volume_by_weekday <stats> width height %}
    Args:
      stats - a stats object containing volume_by_day_of_week
    """
    volmap = [0] * 7
    vols = stats.get('volume_by_day_of_week', {})
    if not volmap:
        raise ChartError('Daily volumes unavailable')

    for weekday, volume_ml in vols.iteritems():
        volmap[int(weekday)] += format_volume(volume_ml, kwargs)[0]
    return _weekday_chart_common(volmap)


def chart_sessions_by_weekday(stats, *args, **kwargs):
    data = stats.get('volume_by_day_of_week', {})
    weekdays = [0] * 7
    for weekday, volume_ml in data.iteritems():
        weekdays[int(weekday)] += format_volume(volume_ml, kwargs)[0]
    return _weekday_chart_common(weekdays)


def chart_sessions_by_volume(stats, *args, **kwargs):
    buckets = [0] * 6
    labels = [
        '<1',
        '1.0-1.9',
        '2.0-2.9',
        '3.0-3.9',
        '4.0-4.9',
        '5+'
    ]
    volmap = stats.get('volume_by_session', {})
    for session_volume in volmap.values():
        volume = round(format_volume(session_volume, kwargs)[0], 1)
        intval = int(volume)
        if intval >= len(buckets):
            buckets[-1] += 1
        else:
            buckets[intval] += 1

    res = {
        'xAxis': {
            'categories': labels,
        },
        'series': [
            {'data': buckets},
        ],
        'yAxis': {
            'min': 0,
        },
        'chart': {
            'defaultSeriesType': 'column',
        }
    }
    return res


def chart_users_by_volume(stats, *args, **kwargs):
    vols = stats.get('volume_by_drinker')
    if not vols:
        raise ChartError('no data')

    data = []
    for username, volume in vols.iteritems():
        if not username:
            username = 'Guest'
        volume, units = format_volume(volume, kwargs)
        label = '<b>%s</b> (%.1f %s)' % (username, volume, units)
        data.append((label, volume))

    other_vol = 0
    for username, pints in data[10:]:
        other_vol += pints

    def _sort_vol_desc(a, b):
        return cmp(b[1], a[1])

    data.sort(_sort_vol_desc)
    data = data[:10]
    data.reverse()

    if other_vol:
        label = '<b>%s</b> (%.1f)' % ('all others', other_vol)
        data.append((label, other_vol))

    res = {
        'series': [
          {
              'type': 'pie',
              'name': 'Drinkers by Volume',
              'data': data,
          }
        ],
        'yAxis': {
            'min': 0,
        },
        'chart': {
            'defaultSeriesType': 'column',
        },
        'tooltip': {
            'enabled': False,
        },
    }
    return res


def _weekday_chart_common(vals):
    labels = ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat']

    # convert from 0=Monday to 0=Sunday
    #vals.insert(0, vals.pop(-1))

    res = {
        'xAxis': {
            'categories': labels,
        },
        'yAxis': {
            'min': 0,
        },
        'series': [
            {'data': vals},
        ],
        'tooltip': {
            'enabled': False,
        },
        'chart': {
            'defaultSeriesType': 'column',
        }
    }
    return res



from django.conf import settings
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.auth.decorators import user_passes_test


def staff_member_required(view_func, redirect_field_name=REDIRECT_FIELD_NAME,
        login_url=settings.KEGBOT_ADMIN_LOGIN_URL):
    """
    Clone of django.contrib.admin.views.decorators.staff_member_required that
    uses `settings.KEGBOT_ADMIN_LOGIN_URL` as the default login URL.
    """
    return user_passes_test(
        lambda u: u.is_active and u.is_staff,
        login_url=login_url,
        redirect_field_name=redirect_field_name
    )(view_func)



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Local Django authentication backend."""

import uuid

from django.db import IntegrityError
from django.contrib.auth.backends import ModelBackend

from pykeg.core import models
from pykeg.web.auth import AuthBackend
from pykeg.web.auth import UserExistsException
from pykeg.util.email import build_message


class LocalAuthBackend(ModelBackend, AuthBackend):

    def is_manager(self, user):
        return user.is_staff

    def set_is_manager(self, user, is_manager):
        if user.is_staff != is_manager:
            user.is_staff = is_manager
            user.save()

    def is_owner(self, user):
        return user.is_superuser

    def set_is_owner(self, user, is_owner):
        if user.is_superuser != is_owner:
            user.is_superuser = is_owner
            user.save()

    def register(self, email, username, password=None, photo=None):
        try:
            user = models.User.objects.create(username=username, email=email)
        except IntegrityError as e:
            raise UserExistsException(e)

        if password:
            user.set_password(password)
        else:
            user.set_unusable_password()
            user.activation_key = str(uuid.uuid4()).replace('-', '')

        if photo:
            pic = models.Picture.objects.create(user=user)
            pic.image.save(photo.name, photo)
            pic.save()
            user.mugshot = pic

        user.save()

        kbsite = models.KegbotSite.get()
        template = None

        if user.activation_key:
            # Needs further activation: send activation e-mail.
            template = 'registration/email_activate_registration.html'
            url = kbsite.reverse_full('activate-account', args=(),
                kwargs={'activation_key': user.activation_key})
        elif email:
            # User already activated, send "complete" email.
            template = 'registration/email_registration_complete.html'
            url = kbsite.reverse_full('kb-account-main', args=(), kwargs={})

        if template:
            context = {
                'site_name': kbsite.title,
                'url': url,
            }
            message = build_message(email, template, context)
            if message:
                message.send(fail_silently=True)

        return user



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from django.contrib import messages
from pykeg.web.decorators import staff_member_required
from django.shortcuts import render_to_response
from django.template import RequestContext

from . import forms


@staff_member_required
def admin_settings(request, plugin):
    context = RequestContext(request)
    settings_form = plugin.get_site_settings_form()

    if request.method == 'POST':
        if 'submit-settings' in request.POST:
            settings_form = forms.SiteSettingsForm(request.POST)
            if settings_form.is_valid():
                plugin.save_site_settings_form(settings_form)
                messages.success(request, 'Settings updated')

    context['plugin'] = plugin
    context['settings_form'] = settings_form

    return render_to_response('contrib/webhook/webhook_admin_settings.html',
            context_instance=context)



# Kegbot local settings, for travis-ci.org build

import os

HOME = os.environ['HOME']

DEBUG = True
TEMPLATE_DEBUG = DEBUG

DATABASES = {'default': {'ENGINE': 'django.db.backends.mysql', 'NAME': 'kegbot_travis_test', 'HOST': '', 'USER': 'root', 'PASSWORD': '', 'OPTIONS': {'init_command': 'SET storage_engine=INNODB'}}}

KEGBOT_ROOT = HOME + '/kegbot-data'

MEDIA_ROOT = KEGBOT_ROOT  + '/media'

STATIC_ROOT = KEGBOT_ROOT + '/static'

SECRET_KEY = 'testkey'

EMAIL_BACKEND = 'django.core.mail.backends.console.EmailBackend'
EMAIL_FROM_ADDRESS = 'no-reply@example.com'



from django import forms
from django.contrib.auth.forms import AuthenticationForm

from pykeg.core import models
ALL_PICTURES = models.Picture.objects.all()


class LoginForm(AuthenticationForm):
    next_page = forms.CharField(required=False, widget=forms.HiddenInput)


class ActivateAccountForm(forms.Form):
    password = forms.CharField(required=True, widget=forms.PasswordInput)
    password2 = forms.CharField(required=True, widget=forms.PasswordInput)

    def clean_password2(self):
        password1 = self.cleaned_data.get("password")
        password2 = self.cleaned_data.get("password2")
        if password1 and password2 and password1 != password2:
            raise forms.ValidationError('Passwords do not match.')
        return password2


class InvitationForm(forms.Form):
    email = forms.EmailField(required=True,
        help_text='E-mail address to invite')


class ProfileForm(forms.Form):
    new_mugshot = forms.ImageField(required=False)
    display_name = forms.CharField(required=True)


class RegenerateApiKeyForm(forms.Form):
    pass


class DeletePictureForm(forms.Form):
    picture = forms.ModelChoiceField(queryset=ALL_PICTURES, required=True,
         widget=forms.HiddenInput)



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

import urlparse

from django import forms
from django.conf import settings
from django.contrib.auth.tokens import default_token_generator
from django.template import loader
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_encode
from django.utils.translation import ugettext_lazy as _

try:
    from django.contrib.auth import get_user_model
    User = get_user_model()
except ImportError:
    from django.contrib.auth.models import User

from pykeg.core import models
from pykeg.backend import get_kegbot_backend


class KegbotRegistrationForm(forms.ModelForm):
    class Meta:
        model = models.User
        fields = ('email', 'username')

    password1 = forms.CharField(widget=forms.PasswordInput,
                                label=_("Password"))
    password2 = forms.CharField(widget=forms.PasswordInput,
                                label=_("Password (again)"))

    def clean(self):
        super(KegbotRegistrationForm, self).clean()
        if 'password1' in self.cleaned_data and 'password2' in self.cleaned_data:
            if self.cleaned_data['password1'] != self.cleaned_data['password2']:
                raise forms.ValidationError(_("The two password fields didn't match."))
        return self.cleaned_data


class PasswordResetForm(forms.Form):
    email = forms.EmailField(label=_("Email"), max_length=254)

    def save(self, domain_override=None,
             subject_template_name='registration/password_reset_subject.txt',
             email_template_name='registration/password_reset_email.html',
             use_https=False, token_generator=default_token_generator,
             from_email=None, request=None, html_email_template_name=None):
        """
        Generates a one-use only link for resetting password and sends to the
        user.
        """
        from django.core.mail import send_mail
        email = self.cleaned_data["email"]
        active_users = User._default_manager.filter(
            email__iexact=email, is_active=True)
        for user in active_users:
            # Make sure that no email is sent to a user that actually has
            # a password marked as unusable
            if not user.has_usable_password():
                continue
            from_email = getattr(settings, 'EMAIL_FROM_ADDRESS', from_email)

            be = get_kegbot_backend()
            base_url = be.get_base_url()
            parsed = urlparse.urlparse(base_url)
            domain = parsed.netloc
            protocol = parsed.scheme

            kbsite = models.KegbotSite.get()
            site_name = kbsite.title
            c = {
                'email': user.email,
                'site_name': site_name,
                'uid': urlsafe_base64_encode(force_bytes(user.pk)),
                'user': user,
                'token': token_generator.make_token(user),
                'domain': domain,
                'protocol': protocol,
            }
            subject = loader.render_to_string(subject_template_name, c)
            # Email subject *must not* contain newlines
            subject = ''.join(subject.splitlines())
            email = loader.render_to_string(email_template_name, c)
            send_mail(subject, email, from_email, [user.email])



from __future__ import absolute_import

import os
from celery import Celery
from django.conf import settings

# set the default Django settings module for the 'celery' program.
os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'pykeg.settings')

app = Celery('pykeg')
app.config_from_object('django.conf:settings')
app.autodiscover_tasks(lambda: settings.INSTALLED_APPS)

plugin_tasks = lambda: ['.'.join(x.split('.')[:-2]) for x in settings.KEGBOT_PLUGINS]
app.autodiscover_tasks(plugin_tasks())



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Unittests for Untappd plugin."""

from mock import patch

from django.test import TransactionTestCase
from django.utils import timezone
from pykeg.contrib.untappd import plugin
from pykeg.contrib.foursquare.plugin import FoursquarePlugin
from pykeg.plugin.datastore import InMemoryDatastore

from pykeg.core import models
from pykeg.backend import get_kegbot_backend


class UntappdTests(TransactionTestCase):

    def setUp(self):
        self.datastore = InMemoryDatastore(plugin.UntappdPlugin.get_short_name())

        self.fsq = FoursquarePlugin(datastore=self.datastore)
        self.fake_plugin_registry = {'foursquare': self.fsq}

        self.plugin = plugin.UntappdPlugin(datastore=self.datastore,
            plugin_registry=self.fake_plugin_registry)
        self.user = models.User.objects.create(username='untappd_test')
        self.backend = get_kegbot_backend()
        self.tap = self.backend.create_tap('Test Tap', 'test.flow0')
        self.keg = self.backend.start_keg(tap=self.tap, beverage_name='Test Beer',
            beverage_type='beer', producer_name='Test Producer', style_name='Test Style')
        self.beverage = self.keg.type
        self.beverage.untappd_beer_id = '9876'
        self.beverage.save()

    def test_plugin(self):
        self.assertEqual((None, None), self.plugin.get_credentials())
        self.assertEqual({}, self.plugin.get_user_profile(self.user))

        fake_profile = {
            'foo': 'bar'
        }
        self.plugin.save_user_profile(self.user, fake_profile)
        self.assertEquals(fake_profile, self.plugin.get_user_profile(self.user))

    def test_drink_poured_no_foursquare(self):
        self.plugin.save_user_token(self.user, 'fake-token')
        self.assertEqual('fake-token', self.plugin.get_user_token(self.user))

        fake_drink = models.Drink.objects.create(keg=self.keg, volume_ml=1000, ticks=1000,
            user=self.user, time=timezone.now(), shout='Hello')
        fake_event = models.SystemEvent.objects.create(kind=models.SystemEvent.DRINK_POURED,
            drink=fake_drink, user=self.user, keg=self.keg, time=fake_drink.time)

        with patch('pykeg.contrib.untappd.tasks.untappd_checkin.delay') as mock_checkin:
            self.plugin.handle_new_events([fake_event])
            mock_checkin.assert_called_with('fake-token', '9876', 'UTC', shout='Hello',
                foursquare_client_id=None,
                foursquare_client_secret=None,
                foursquare_venue_id=None)

    def test_drink_poured_with_foursquare(self):
        fsq_settings = self.fsq.get_site_settings_form()
        fsq_settings.cleaned_data = {
            'venue_id': '54321',
            'client_id': 'fake-client-id',
            'client_secret': 'fake-client-secret',
        }
        self.fsq.save_site_settings_form(fsq_settings)

        self.plugin.save_user_token(self.user, 'fake-token')
        self.assertEqual('fake-token', self.plugin.get_user_token(self.user))

        fake_drink = models.Drink.objects.create(keg=self.keg, volume_ml=1000, ticks=1000,
            user=self.user, time=timezone.now(), shout='Hello2')
        fake_event = models.SystemEvent.objects.create(kind=models.SystemEvent.DRINK_POURED,
            drink=fake_drink, user=self.user, keg=self.keg, time=fake_drink.time)

        with patch('pykeg.contrib.untappd.tasks.untappd_checkin.delay') as mock_checkin:
            self.plugin.handle_new_events([fake_event])
            mock_checkin.assert_called_with('fake-token', '9876', 'UTC', shout='Hello2',
                foursquare_client_id='fake-client-id',
                foursquare_client_secret='fake-client-secret',
                foursquare_venue_id='54321')



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Methods to generate cached statistics from drinks."""

import copy
import inspect
import logging
import pytz

from django.utils.timezone import localtime
from pykeg.core import models
from kegbot.util import util

STAT_MAP = {}

logger = logging.getLogger(__name__)


class StatsView:
    def __init__(self, user=None, session=None, keg=None):
        self.user = user
        self.session = session
        self.keg = keg

    def __str__(self):
        ret = 'view: '
        if not self.user and not self.keg and not self.session:
            ret += 'system'
        if self.user:
            ret += 'user={} '.format(self.user.username)
        if self.session:
            ret += 'session={} '.format(self.session.id)
        if self.keg:
            ret += 'keg={}'.format(self.keg.id)
        return ret

    def as_tuple(self):
        return (
            self.user.id if self.user else None,
            self.session.id if self.session else None,
            self.keg.id if self.keg else None
        )

    def get_prior_drinks(self, drink):
        """Returns the queryset of prior drinks occurring in this view.

        The result can be empty, which implies the drink is the first drink
        for the given {user, session, keg}.
        """
        qs = models.Drink.objects.filter(id__lt=drink.id)
        if self.user:
            qs = qs.filter(user=self.user.id)
        if self.session:
            qs = qs.filter(session=self.session.id)
        if self.keg:
            qs = qs.filter(keg=self.keg.id)
        return qs.order_by('-id')


class StatsBuilder:
    """Derives statistics from drinks."""

    def __init__(self):
        self.functions = []
        for name, fn in inspect.getmembers(self, inspect.ismethod):
            if not name.startswith('_') and name != 'build':
                self.functions.append((name, fn))

    def build(self, drink, previous_stats):
        if previous_stats is None:
            previous_stats = util.AttrDict()

        logger.debug('build: drink={}'.format(drink.id))
        stats = util.AttrDict()

        for statname, fn in self.functions:
            previous_value = previous_stats.get(statname, None)
            if previous_value is not None:
                val = fn(drink, previous_stats, previous_value=previous_value)
            else:
                val = fn(drink, previous_stats)

            if val is None:
                raise ValueError('Stat generator for %s returned None' % statname)
            stats[statname] = val
        return stats

    def last_drink_id(self, drink, previous_stats, previous_value=0):
        return drink.id

    def keg_ids(self, drink, previous_stats, previous_value=[]):
        ret = copy.copy(previous_value)
        if drink.keg.id not in ret:
            ret.append(drink.keg.id)
        return ret

    def total_volume_ml(self, drink, previous_stats, previous_value=0):
        return previous_value + drink.volume_ml

    def total_pours(self, drink, previous_stats, previous_value=0):
        return previous_value + 1

    def average_volume_ml(self, drink, previous_stats, previous_value=0):
        vol = previous_stats.get('total_volume_ml', 0)
        count = previous_stats.get('total_pours', 0)
        vol += drink.volume_ml
        count += 1
        return vol / float(count)

    def greatest_volume_ml(self, drink, previous_stats, previous_value=0):
        return float(max(drink.volume_ml, previous_value))

    def greatest_volume_id(self, drink, previous_stats, previous_value=0):
        if drink.volume_ml > previous_stats.get('greatest_volume_ml', 0):
            return drink.id
        return previous_value

    def volume_by_day_of_week(self, drink, previous_stats, previous_value={}):
        ret = copy.copy(previous_value)
        tz = pytz.timezone(drink.session.timezone)
        local_time = localtime(drink.session.start_time, timezone=tz)
        drink_weekday = str(local_time.strftime('%w'))
        ret[drink_weekday] = ret.get(drink_weekday, 0) + drink.volume_ml
        return ret

    def registered_drinkers(self, drink, previous_stats, previous_value=[]):
        ret = copy.copy(previous_value)
        if drink.user:
            user_id = str(drink.user.id)
            if user_id not in ret:
                ret.append(user_id)
        return ret

    def sessions_count(self, drink, previous_stats, previous_value=0):
        # Use volume_by_session, ensuring our session is captured
        # by injecting a dummy value.
        prev_sessions = previous_stats.get('volume_by_session', {})
        ret = len(prev_sessions.keys())
        if str(drink.session.id) not in prev_sessions:
            ret += 1
        return ret

    def volume_by_year(self, drink, previous_stats, previous_value={}):
        ret = copy.copy(previous_value)
        year = str(drink.time.year)
        ret[year] = ret.get(year, 0) + drink.volume_ml
        return ret

    def has_guest_pour(self, drink, previous_stats, previous_value=False):
        if previous_value:
            return True
        return drink.is_guest_pour()

    def volume_by_drinker(self, drink, previous_stats, previous_value={}):
        ret = copy.copy(previous_value)
        user_id = str(drink.user.id)
        orig = ret.get(user_id, 0)
        ret[user_id] = float(orig + drink.volume_ml)
        return ret

    def volume_by_session(self, drink, previous_stats, previous_value={}):
        ret = copy.copy(previous_value)
        session_id = str(drink.session.id)
        ret[session_id] = ret.get(session_id, 0) + drink.volume_ml
        return ret

    def largest_session(self, drink, previous_stats, previous_value={}):
        if drink.session.volume_ml >= previous_value.get('volume_ml', 0):
            return {
                'session_id': drink.session.id,
                'volume_ml': drink.session.volume_ml,
            }
        return previous_value

BUILDER = StatsBuilder()

### Public methods


def invalidate(drink_id):
    """Clears all statistics starting from (and including) drink_id."""
    logger.debug('--- Invalidating stats since id {}'.format(drink_id))
    models.Stats.objects.filter(drink_id__gte=drink_id).delete()


def invalidate_all():
    logger.debug('--- Invalidating ALL stats')
    models.Stats.objects.all().delete()


def build_for_id(drink_id):
    """Build stats up to and including this drink id, *without*
    considering any subsequent drinks.
    """
    try:
        drink = models.Drink.objects.get(pk=drink_id)
    except models.Drink.DoesNotExist:
        logger.warning('No drink exists with id={}'.format(drink_id))
        return
    _build_all_views(drink)


def rebuild_from_id(drink_id, cb=None):
    """Builds statistics stating from `drink_id`.  Unlike `build_for_id()`,
    this method computes statistics for any subsequent drinks found in the
    database.
    """
    invalidate(drink_id)
    results_cache = {}
    for drink in models.Drink.objects.filter(id__gte=drink_id).order_by('id'):
        _build_all_views(drink, results_cache=results_cache)
        if cb:
            cb(results_cache)


def _build_single_view(drink, view, prior_stats=None):
    """Generates all stats for a drink in the specified view.

    Args:
        drink: The target drink.
        view: The view.
        prior_stats: Dictionary of previous stats for this view.  If None,
            indicates that the prior stats are unknown.  When prior stats
            are unknown, they will be queried or generated as needed.
    """
    logger.debug('>>> Building stats for {}'.format(view))

    build_list = [drink]
    if prior_stats is None:
        prior_stats = util.AttrDict()
        prior_drinks_in_view = view.get_prior_drinks(drink)
        if prior_drinks_in_view.count():
            # Starting with the most recent prior drink, get its stats row.
            # If stats don't exist, add drink to build list and continue
            # until a stats row is found or all drinks are exhausted.
            # N.B. we avoid a recursive algorithm due to potentially large recursion
            # depth in certain cases.
            for prior_drink in prior_drinks_in_view:
                try:
                    prior_stats = util.AttrDict(
                        models.Stats.objects.get(drink=prior_drink, user=view.user,
                            session=view.session, keg=view.keg).stats
                    )
                    break
                except models.Stats.DoesNotExist:
                    build_list.insert(0, prior_drink)

    # Build all drinks on the hit list.
    for build_drink in build_list:
        logger.debug('  - operating on drink {}'.format(build_drink.id))
        stats = BUILDER.build(drink=build_drink, previous_stats=prior_stats)
        models.Stats.objects.create(drink=build_drink, user=view.user, time=build_drink.time,
            session=view.session, keg=view.keg, stats=stats, is_first=(not prior_stats))
        prior_stats = stats
    logger.debug('<<< Done.')
    return stats


def _build_all_views(drink, results_cache=None):
    """Generates all stats "views" based on `drink`.

    Args:
        drink: The target drink.
        invalidate_first: If True, all statistics starting for and following
            this drink will be deleted first.  Typically this is True when the
            target drink was modified, and False when it is a new (newest) Drink.
        results_cache: A cache mapping view.as_tuple() -> stats.  Will be modified
            with the results of this method.
    """
    if results_cache is None:
        results_cache = {}

    for user in (None, drink.user):
        for session in (None, drink.session):
            for keg in (None, drink.keg):
                view = StatsView(user, session, keg)
                cache_key = view.as_tuple()

                prior_stats = results_cache.get(cache_key)
                stats = _build_single_view(drink, view, prior_stats=prior_stats)
                results_cache[cache_key] = stats



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Custom signals from the Kegbot backend."""

from django.dispatch import Signal

user_created = Signal(providing_args=['user'])

tap_created = Signal(providing_args=['tap'])

auth_token_created = Signal(providing_args=['token'])

drink_recorded = Signal(providing_args=['drink'])

drink_canceled = Signal(providing_args=['drink'])

drink_assigned = Signal(providing_args=['drink', 'previous_user'])

drink_adjusted = Signal(providing_args=['drink', 'previous_volume'])

temperature_recorded = Signal(providing_args=['record'])

keg_created = Signal(providing_args=['keg'])

keg_attached = Signal(providing_args=['keg', 'tap'])

keg_ended = Signal(providing_args=['keg'])



from __future__ import absolute_import

from jsonfield import JSONField as BaseJSONField


class JSONField(BaseJSONField):
    pass

try:
    from south.modelsinspector import add_introspection_rules
    add_introspection_rules([], ["^pykeg.core.jsonfield\.JSONField"])
except ImportError:
    pass



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from django.core.management.base import NoArgsCommand

from pykeg.core import models
from pykeg.core.management.commands.common import progbar


class Command(NoArgsCommand):
    help = u'Regenerate all system events.'
    args = '<none>'

    def handle(self, **options):
        events = models.SystemEvent.objects.all()
        num_events = len(events)
        progbar('clear events', 0, num_events)
        events.delete()
        progbar('clear events', num_events, num_events)
        print ''

        pos = 0
        drinks = models.Drink.objects.all()
        count = drinks.count()
        for d in drinks.order_by('time'):
            pos += 1
            progbar('create new events', pos, count)
            models.SystemEvent.build_events_for_drink(d)
        print ''

        print 'done!'



from django import forms
from pykeg.core import models


class NotificationSettingsForm(forms.ModelForm):
    class Meta:
        model = models.NotificationSettings
        exclude = ['user', 'backend']



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from django.contrib.auth import authenticate
from django.contrib.auth import login
from django.shortcuts import redirect
from django.shortcuts import render_to_response
from django.template import RequestContext

from pykeg.web.kbregistration.forms import KegbotRegistrationForm
from pykeg.backend import get_kegbot_backend

from pykeg.core import models

"""Kegbot-aware registration views."""


def register(request):
    context = RequestContext(request)
    form = KegbotRegistrationForm()

    # Check if we need an invitation before processing the request further.
    invite = None
    if request.kbsite.registration_mode != 'public':
        invite_code = None
        if 'invite_code' in request.GET:
            invite_code = request.GET['invite_code']
            request.session['invite_code'] = invite_code
        else:
            invite_code = request.session.get('invite_code', None)

        if not invite_code:
            r = render_to_response('registration/invitation_required.html',
                context_instance=context)
            r.status_code = 401
            return r

        try:
            invite = models.Invitation.objects.get(invite_code=invite_code)
        except models.Invitation.DoesNotExist:
            pass

        if not invite or invite.is_expired():
            r = render_to_response('registration/invitation_expired.html',
                context_instance=context)
            r.status_code = 401
            return r

    if request.method == 'POST':
        form = KegbotRegistrationForm(request.POST)
        if form.is_valid():
            username = form.cleaned_data['username']
            email = form.cleaned_data['email']
            password = form.cleaned_data.get('password1')

            backend = get_kegbot_backend()
            backend.create_new_user(username=username, email=email, password=password)

            if invite:
                invite.delete()
                if 'invite_code' in request.session:
                    del request.session['invite_code']

            if password:
                new_user = authenticate(username=username, password=password)
                login(request, new_user)
                return redirect('kb-account-main')

            return render_to_response('registration/registration_complete.html',
                context_instance=context)

    context['form'] = form
    return render_to_response('registration/registration_form.html',
        context_instance=context)
